
#include <csignal>
#include <cstdio>
#include "armemul.h"
#include "decode.h"

#if HAVE_SYS_RESOURCE_H && HAVE_SYS_TIME_H
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#endif

#include "auto_impl.h"
#include "arm_dec.h"

#define DEF_CLASS(a,b) b,
const char *mnemonic_table[] = {
#include "inst_class.h"
};

static const UInt16 cond_table[] =
{0xf0f0, 0x0f0f, 0xcccc, 0x3333,
 0xff00, 0x00ff, 0xaaaa, 0x5555,
 0x0c0c, 0xf3f3, 0xaa55, 0x55aa,
 0x0a05, 0xf5fa, 0xffff, 0x0000};

using namespace emulator;

static inline UInt32 eval_cond(arm_emulator *emu, arm_inst_t inst)
{
	UInt32 val;
	val = (cond_table[COND] >> (CPSR>>28))&1;
#if 0
	switch (COND) {
		case 0: val = Z_FLAG; break;
		case 1: val = !Z_FLAG; break;
		case 2: val = C_FLAG; break;
		case 3: val = !C_FLAG; break;
		case 4: val = N_FLAG; break;
		case 5: val = !N_FLAG; break;
		case 6: val = V_FLAG; break;
		case 7: val = !V_FLAG; break;
		case 8: val = C_FLAG & ~Z_FLAG; break;
		case 9: val = !(C_FLAG & ~Z_FLAG); break;
		case 10: val = N_FLAG == V_FLAG; break;
		case 11: val = N_FLAG ^ V_FLAG; break;
		case 12: val = !(Z_FLAG | (N_FLAG ^ V_FLAG)); break;
		case 13: val = Z_FLAG | (N_FLAG ^ V_FLAG); break;
		case 14: val = 1; break;
		default: val = 0; break;
	}
#endif
	return val;
}

arm_emulator::arm_emulator(bool verbose, bool emulate_syscall) :
	verbose(verbose), emulate_syscall(emulate_syscall), in_fpe(false)
{
	init_registers();
	reset_instruction_counters();
	init_queues();
}

arm_emulator::~arm_emulator()
{
}


void arm_emulator::execute(arm_inst_t inst, word_t addr)
{

#if 0
	{
		int x=0,i;
		for (i=0;i<15;i++) x=x+my_regs.gpr[i];
			fprintf(stderr, "0x%x 0x%x\n", PC,x);
	}
	arm_disassemble_to_stream(inst, get_pc(), stderr);
#endif

	clear_pc_update();

	/*operation sees PC+8*/
	write_gpr2(PCIND, addr+8);

	if (eval_cond(this, inst)) 
		decode_main(this, inst);
	else
		ncount++;

	icount++;
	/*if pc not updated, increment by 4*/
	if (!is_pc_updated()) write_gpr2(PCIND, addr+4);

}

void arm_emulator::run(UInt64 count)
{
	arm_inst_t inst;

#if HAVE_SYS_RESOURCE_H && HAVE_SYS_TIME_H
	struct timeval begin_u, end_u, begin_s, end_s;
	struct rusage usg;
	float user_time, sys_time;
	getrusage(RUSAGE_SELF, &usg);
	begin_u = usg.ru_utime;
	begin_s = usg.ru_stime;
#endif

	running = true;
	while(running && icount<count)
	{
		inst = fetch_inst(get_pc());
		execute(inst, get_pc());
	}

	counters[INST_FPE] /= 2;
	dump_instruction_counters(stderr);
	fprintf(stderr, "Total memory pages allocated : %d\n",
			mem.get_page_count());

#if HAVE_SYS_RESOURCE_H && HAVE_SYS_TIME_H
	getrusage(RUSAGE_SELF, &usg);
	end_u = usg.ru_utime;
	end_s = usg.ru_stime;
	user_time = (end_u.tv_sec+end_u.tv_usec/1000000.0)-
		(begin_u.tv_sec+begin_u.tv_usec/1000000.0);
	sys_time = (end_s.tv_sec+end_s.tv_usec/1000000.0)-
		(begin_s.tv_sec+begin_s.tv_usec/1000000.0);
	fprintf(stderr, "Total user time  : %.3f sec.\n"
		"Total system time: %.3f sec.\n"
		"Simulation speed : %.3e inst/sec.\n",
		user_time, sys_time, icount/(user_time+sys_time));
#endif
}

void arm_emulator::init_queues()
{
	input_queue = new std::queue<ns_data>[16];
}

void arm_emulator::init_registers()
{
	for (int i=0; i<16; i++)
		my_regs.gpr[i] = 0;

	my_regs.cpsr = 0;
	my_regs.spsr = 0;
	my_regs.pcflag = 0;
}

void arm_emulator::reset_instruction_counters()
{
	int i;
	icount = 0;
	ncount = 0;
	for(i = 0; i < INST_TOTAL; i++)
		counters[i] = 0;
}

void arm_emulator::reset()
{
	init_registers();
	mem.reset();
	reset_instruction_counters();
}

void arm_emulator::dump_instruction_counters(FILE *fp)
{
	int i;
	
	fprintf(fp, "\n");
	for(i = 0; i < INST_TOTAL; i++) {
		if(counters[i] > 0) {
			fprintf(fp, "%s : ", mnemonic_table[i]);
			dump_int64(counters[i], fp);
			if(counters[i] > 1000000) {
				fputs(" (", fp);
				dump_int64_smart(counters[i], fp);
				fputs(")", fp);
			}
			fputs("\n", fp);
		}
	}
	fputs("Total instructions : ", fp);
	dump_int64(icount, fp);
	fputs(" (", fp);
	dump_int64_smart(icount, fp);
	fputs(")", fp);
	fputs(" including ", fp);
	dump_int64(ncount, fp);
	fputs(" (", fp);
	dump_int64_smart(ncount, fp);
	fputs(")", fp);
	fputs(" nullified\n", fp);
}
