/*************************************************************************
    Copyright (C) 2002,2003 Wei Qin
    See file COPYING for more information.

    This program is free software; you can redistribute it and/or modify    
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
*************************************************************************/

#include <cstdio>
#include "misc.h"
#include "armemul.h"
#include "dpi.h"

static const char *arm_dpinames[] = 
{"and", "eor", "sub", "rsb", "add", "adc", "sbc", "rsc",
 "tst", "teq", "cmp", "cmn", "orr", "mov", "bic", "mvn"};

using namespace emulator;

/*utilities for alu operations*/
static inline UInt32 rotate_right(UInt32 val, UInt32 imm)
{
	return (val >> imm ) | (val << (32 - imm));
}

/*
  Shifter_operand1 and shifter_operand1_s actually was in one before.
  So does shifter_operand2's and shifter_oeprand3's.
  Separating them helps simulation speed a little bit (8.31->8.49Mips)
*/

/*immediate shifter operand*/
static inline UInt32 shifter_operand1(armulator *emu, arm_inst_t inst)
{
	UInt32 imm = inst & 0xff;
	UInt32 shift = (inst>>7) & 0x1e;
	UInt32 result;

	result = rotate_right(imm, shift);
	return result;
}

/*immediate shifter operand*/
static inline UInt32 shifter_operand1_s(armulator *emu, arm_inst_t inst)
{
	UInt32 imm = inst & 0xff;
	UInt32 shift = (inst>>7) & 0x1e;
	UInt32 result;

	result = rotate_right(imm, shift);

	if (!shift)
		emu->shifter_carry = C_FLAG;
	else
		emu->shifter_carry = BIT31(result);
	return result;
}

/*shift imediate shifter operand*/
static inline UInt32 shifter_operand2(armulator *emu, arm_inst_t inst)
{
	UInt32 shift_imm = (inst>>7) & 0x01f, val = RM, result;
	UInt8 type = (inst>>5) & 0x03;

	switch (type) {
		case 0: /*LSL*/	
			result = val << shift_imm;
			break;
		case 1: /*LSR*/
			if (shift_imm) {
				result = val >> shift_imm;
			} else {
				result = 0;
			}
			break;
		case 2: /*ASR*/
			if (shift_imm) {
				result = (SInt32)val>>shift_imm;
			}
			else {
				result = BIT31(val)?~0:BIT31(val);
			}
			break;
		default: /*ROR*/
			if (shift_imm) {
				result = rotate_right(val, shift_imm);
			}
			else { /*RRX*/
				result = (val>>1) |	(C_FLAG<<31);
			}
			break;
			
	}
	return result;
}

/*shift imediate shifter operand*/
static inline UInt32 shifter_operand2_s(armulator *emu, arm_inst_t inst)
{
	UInt32 shift_imm = (inst>>7) & 0x01f, val = RM, result;
	UInt8 type = (inst>>5) & 0x03;

	switch (type) {
		case 0: /*LSL*/	
			result = val << shift_imm;
			emu->shifter_carry = shift_imm?BITn(val,32-shift_imm):C_FLAG;
			break;
		case 1: /*LSR*/
			if (shift_imm) {
				result = val >> shift_imm;
				emu->shifter_carry = BITn(val, shift_imm-1);
			} else {
				result = 0;
				emu->shifter_carry = BIT31(val);
			}
			break;
		case 2: /*ASR*/
			if (shift_imm) {
				result = (SInt32)val>>shift_imm;
				emu->shifter_carry = BITn(val,shift_imm-1);
			}
			else {
				result = BIT31(val)?~0:BIT31(val);
				emu->shifter_carry = BIT31(val);
			}
			break;
		default: /*ROR*/
			if (shift_imm) {
				result = rotate_right(val, shift_imm);
				emu->shifter_carry = BITn(val, shift_imm-1);
			}
			else { /*RRX*/
				result = (val>>1) |	(C_FLAG<<31);
				emu->shifter_carry = BIT0(val);
			}
			break;
	}
	return result;
}

/*shift register shifter operand*/
static inline UInt32 shifter_operand3(armulator *emu, arm_inst_t inst)
{
	UInt32 shift_imm = RS & 0xff, val = RM, result;
	UInt8 type = (inst>>5) & 0x03;

	switch (type) {
		case 0: /*LSL*/	
			if (shift_imm==0) {
				result = val;
			} else if (shift_imm<32) {
				result = val << shift_imm;
			} else if (shift_imm==32) {
				result = 0;
			} else {
				result = 0;
			}
			break;
		case 1: /*LSR*/
			if (shift_imm==0) {
				result = val;
			} else if (shift_imm<32) {
				result = val >> shift_imm;
			} else if (shift_imm==32) {
				result = 0;
			} else {
				result = 0;
			}
			break;
		case 2: /*ASR*/
			if (shift_imm==0) {
				result = val;
			} else if (shift_imm<32) {
				result = (SInt32)val >> shift_imm;
			} else if (BIT31(val)) {
				result = ~0;
			} else {
				result = 0;
			}
			break;
		default: /*ROR*/
			if (shift_imm==0) {
				result = val;
			}
			else if ((shift_imm&0x1f)==0) {
				result = val;
			}
			else {
				result = rotate_right(val, shift_imm&0x1f);
			}
			break;
	}
	return result;
}

/*shift register shifter operand*/
static inline UInt32 shifter_operand3_s(armulator *emu, arm_inst_t inst)
{
	UInt32 shift_imm = RS & 0xff, val = RM, result;
	UInt8 type = (inst>>5) & 0x03;

	switch (type) {
		case 0: /*LSL*/	
			if (shift_imm==0) {
				result = val;
				emu->shifter_carry = C_FLAG;
			} else if (shift_imm<32) {
				result = val << shift_imm;
				emu->shifter_carry = BITn(val,32-shift_imm);
			} else if (shift_imm==32) {
				result = 0;
				emu->shifter_carry = BIT0(val);
			} else {
				result = 0;
				emu->shifter_carry = 0;
			}
			break;
		case 1: /*LSR*/
			if (shift_imm==0) {
				result = val;
				emu->shifter_carry = C_FLAG;
			} else if (shift_imm<32) {
				result = val >> shift_imm;
				emu->shifter_carry = BITn(val,shift_imm-1);
			} else if (shift_imm==32) {
				result = 0;
				emu->shifter_carry = BIT31(val);
			} else {
				result = 0;
				emu->shifter_carry = 0;
			}
			break;
		case 2: /*ASR*/
			if (shift_imm==0) {
				result = val;
				emu->shifter_carry = C_FLAG;
			} else if (shift_imm<32) {
				result = (SInt32)val >> shift_imm;
				emu->shifter_carry = BITn(val,shift_imm-1);
			} else if (BIT31(val)) {
				result = ~0;
				emu->shifter_carry = BIT31(val);
			} else {
				result = 0;
				emu->shifter_carry = BIT31(val);
			}
			break;
		default: /*ROR*/
			if (shift_imm==0) {
				result = val;
				emu->shifter_carry = C_FLAG;
			}
			else if ((shift_imm&0x1f)==0) {
				result = val;
				emu->shifter_carry = BIT31(val);
			}
			else {
				result = rotate_right(val, shift_imm&0x1f);
				emu->shifter_carry = BITn(val, (shift_imm&0x1f)-1);
			}
			break;
	}
	return result;
}

/*data processing instructions*/

/*move*/

static inline void set_move_flags(armulator *emu, UInt32 val)
{
#if 0
	ASGN_N(BIT31(val));
	ASGN_Z(val==0);
	ASGN_C(emu->shifter_carry);
#endif
	ASGN_CONDS(ASM_CONDS(BIT31(val), val==0, emu->shifter_carry, V_FLAG));
}

void impl_mov_1(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, shifter_operand1(emu, inst));

	EMULATOR_STUB(mov_1,inst);
}

void impl_mov_1s(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, shifter_operand1_s(emu, inst));
	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_move_flags(emu, RD);

	EMULATOR_STUB(mov_1s,inst);
}

void impl_mov_2(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, shifter_operand2(emu, inst));

	EMULATOR_STUB(mov_2,inst);
}

void impl_mov_2s(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, shifter_operand2_s(emu, inst));
	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_move_flags(emu, RD);

	EMULATOR_STUB(mov_2s,inst);
}

void impl_mov_3(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, shifter_operand3(emu, inst));

	EMULATOR_STUB(mov_3,inst);
}

void impl_mov_3s(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, shifter_operand3_s(emu, inst));
	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_move_flags(emu, RD);

	EMULATOR_STUB(mov_3s,inst);
}

/*MVN*/
void impl_mvn_1(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, ~shifter_operand1(emu, inst));

	EMULATOR_STUB(mvn_1,inst);
}

void impl_mvn_1s(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, ~shifter_operand1_s(emu, inst));
	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_move_flags(emu, RD);

	EMULATOR_STUB(mvn_1s,inst);
}

void impl_mvn_2(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, ~shifter_operand2(emu, inst));

	EMULATOR_STUB(mvn_2,inst);
}

void impl_mvn_2s(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, ~shifter_operand2_s(emu, inst));
	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_move_flags(emu, RD);

	EMULATOR_STUB(mvn_2s,inst);
}

void impl_mvn_3(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, ~shifter_operand3(emu, inst));

	EMULATOR_STUB(mvn_3,inst);
}

void impl_mvn_3s(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, ~shifter_operand3_s(emu, inst));
	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else
		set_move_flags(emu, RD);

	EMULATOR_STUB(mvn_3s,inst);
}

/*ADD*/

static inline void set_add_flags(armulator *emu, UInt64 result)
{
	UInt32 l_res = result, u_res = result>>32;
#if 0
	ASGN_N(BIT31(l_res));
	ASGN_Z(l_res==0);
	ASGN_C(emu->alu_carry);
	ASGN_V(BIT0(u_res)!=BIT31(l_res));
#endif
	ASGN_CONDS(ASM_CONDS(BIT31(l_res), l_res==0, emu->alu_carry,
		BIT0(u_res)!=BIT31(l_res)));
}

static inline UInt64 impl_add64(armulator *emu, SInt32 a, SInt32 b)
{
	UInt64 lhs = a, rhs = b;
	UInt64 result = lhs+rhs;

	emu->alu_carry = ((UInt32)a>>31) ^ ((UInt32)b>>31);
	emu->alu_carry ^= (((UInt32)(result>>32))&0x1);

	return result;
}

static inline UInt64 impl_add64_w_carry(armulator *emu, 
	SInt32 a, SInt32 b, UInt32 c)
{
	UInt64 lhs = a, rhs = b;
	UInt64 result = lhs+rhs+c;

	emu->alu_carry = ((UInt32)a>>31) ^ ((UInt32)b>>31);
	emu->alu_carry ^= (((UInt32)(result>>32))&0x1);

	return result;
}

void impl_add_1(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN + shifter_operand1(emu, inst));

	EMULATOR_STUB(add_1,inst);
}

void impl_add_1s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_add64(emu, RN, shifter_operand1(emu, inst));
	WRITE_REG(RDFLD, (UInt32)result);

	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_add_flags(emu, result);

	EMULATOR_STUB(add_1s,inst);
}

void impl_add_2(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN + shifter_operand2(emu, inst));

	EMULATOR_STUB(add_2,inst);
}

void impl_add_2s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_add64(emu, RN, shifter_operand2(emu, inst));
	WRITE_REG(RDFLD, (UInt32)result);

	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_add_flags(emu, result);

	EMULATOR_STUB(add_2s,inst);
}

void impl_add_3(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN + shifter_operand3(emu, inst));

	EMULATOR_STUB(add_3,inst);
}

void impl_add_3s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_add64(emu, RN, shifter_operand3(emu, inst));
	WRITE_REG(RDFLD, (UInt32)result);

	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else
		set_add_flags(emu, result);

	EMULATOR_STUB(add_3s,inst);
}

/*ADC*/
void impl_adc_1(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN + shifter_operand1(emu, inst) + C_FLAG);

	EMULATOR_STUB(adc_1,inst);
}

void impl_adc_1s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_add64_w_carry(emu, RN, shifter_operand1(emu, inst), C_FLAG);
	WRITE_REG(RDFLD, (UInt32)result);

	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_add_flags(emu, result);

	EMULATOR_STUB(adc_1s,inst);
}

void impl_adc_2(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN + shifter_operand2(emu, inst) + C_FLAG);

	EMULATOR_STUB(adc_2,inst);
}

void impl_adc_2s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_add64_w_carry(emu, RN, shifter_operand2(emu, inst), C_FLAG);
	WRITE_REG(RDFLD, (UInt32)result);

	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_add_flags(emu, result);

	EMULATOR_STUB(adc_2s,inst);
}

void impl_adc_3(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN + shifter_operand3(emu, inst) + C_FLAG);

	EMULATOR_STUB(adc_3,inst);
}

void impl_adc_3s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_add64_w_carry(emu, RN, shifter_operand3(emu, inst), C_FLAG);
	WRITE_REG(RDFLD, (UInt32)result);

	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else
		set_add_flags(emu, result);

	EMULATOR_STUB(adc_3s,inst);
}

/*SUB*/
void set_sub_flags(armulator *emu, UInt64 result)
{
	UInt32 l_res = result, u_res = result>>32;
#if 0
	ASGN_N(BIT31(l_res));
	ASGN_Z(l_res==0);
	ASGN_C(!emu->alu_carry);
	ASGN_V(BIT0(u_res)!=BIT31(l_res));
#endif
	ASGN_CONDS(ASM_CONDS(BIT31(l_res), l_res==0, !emu->alu_carry,
		BIT0(u_res)!=BIT31(l_res)));
}

static inline UInt64 impl_sub64(armulator *emu, SInt32 a, SInt32 b)
{
	UInt64 lhs = a, rhs = b;
	UInt64 result = lhs - rhs;

	emu->alu_carry = ((UInt32)a>>31) ^ ((UInt32)b>>31);
	emu->alu_carry ^= (((UInt32)(result>>32))&0x1);

	return result;
}

static inline UInt64 impl_sub64_w_carry(armulator *emu, SInt32 a, SInt32 b, UInt32 c)
{
	UInt64 lhs = a, rhs = b;
	UInt64 result = lhs - rhs - c;

	emu->alu_carry = ((UInt32)a>>31) ^ ((UInt32)b>>31);
	emu->alu_carry ^= (((UInt32)(result>>32))&0x1);

	return result;
}

void impl_sub_1(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN - shifter_operand1(emu, inst));

	EMULATOR_STUB(sub_1,inst);
}

void impl_sub_1s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_sub64(emu, RN, shifter_operand1(emu, inst));
	WRITE_REG(RDFLD, (UInt32)result);

	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_sub_flags(emu, result);

	EMULATOR_STUB(sub_1s,inst);
}

void impl_sub_2(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN - shifter_operand2(emu, inst));

	EMULATOR_STUB(sub_2,inst);
}

void impl_sub_2s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_sub64(emu, RN, shifter_operand2(emu, inst));
	WRITE_REG(RDFLD, (UInt32)result);

	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_sub_flags(emu, result);

	EMULATOR_STUB(sub_2s,inst);
}

void impl_sub_3(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN - shifter_operand3(emu, inst));

	EMULATOR_STUB(sub_3,inst);
}

void impl_sub_3s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_sub64(emu, RN, shifter_operand3(emu, inst));
	WRITE_REG(RDFLD, (UInt32)result);

	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else
		set_sub_flags(emu, result);

	EMULATOR_STUB(sub_3s,inst);
}

/*SBC*/
void impl_sbc_1(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN - shifter_operand1(emu, inst) - (1-C_FLAG));

	EMULATOR_STUB(sbc_1,inst);
}

void impl_sbc_1s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_sub64_w_carry(emu, RN, shifter_operand1(emu, inst), 1-C_FLAG);
	WRITE_REG(RDFLD, (UInt32)result);

	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_sub_flags(emu, result);

	EMULATOR_STUB(sbc_1s,inst);
}

void impl_sbc_2(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN - shifter_operand2(emu, inst) - (1-C_FLAG));

	EMULATOR_STUB(sbc_2,inst);
}

void impl_sbc_2s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_sub64_w_carry(emu, RN, shifter_operand2(emu, inst), 1-C_FLAG);
	WRITE_REG(RDFLD, (UInt32)result);

	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_sub_flags(emu, result);

	EMULATOR_STUB(sbc_2s,inst);
}

void impl_sbc_3(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN - shifter_operand3(emu, inst) - (1-C_FLAG));

	EMULATOR_STUB(sbc_3,inst);
}

void impl_sbc_3s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_sub64_w_carry(emu, RN, shifter_operand3(emu, inst), 1-C_FLAG);
	WRITE_REG(RDFLD, (UInt32)result);

	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else
		set_sub_flags(emu, result);

	EMULATOR_STUB(sbc_3s,inst);
}

/*RSB*/

void impl_rsb_1(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, shifter_operand1(emu, inst) - RN);

	EMULATOR_STUB(rsb_1,inst);
}

void impl_rsb_1s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_sub64(emu, shifter_operand1(emu, inst), RN);
	WRITE_REG(RDFLD, (UInt32)result);

	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_sub_flags(emu, result);

	EMULATOR_STUB(rsb_1s,inst);
}

void impl_rsb_2(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, shifter_operand2(emu, inst) - RN);

	EMULATOR_STUB(rsb_2,inst);
}

void impl_rsb_2s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_sub64(emu, shifter_operand2(emu, inst), RN);
	WRITE_REG(RDFLD, (UInt32)result);

	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_sub_flags(emu, result);

	EMULATOR_STUB(rsb_2s,inst);
}

void impl_rsb_3(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN - shifter_operand3(emu, inst));

	EMULATOR_STUB(rsb_3,inst);
}

void impl_rsb_3s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_sub64(emu, shifter_operand3(emu, inst), RN);
	WRITE_REG(RDFLD, (UInt32)result);

	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else
		set_sub_flags(emu, result);

	EMULATOR_STUB(rsb_3s,inst);
}

/*RSC*/
void impl_rsc_1(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, shifter_operand1(emu, inst) - RN - (1-C_FLAG));

	EMULATOR_STUB(rsc_1,inst);
}

void impl_rsc_1s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_sub64_w_carry(emu, shifter_operand1(emu, inst), RN, 1-C_FLAG);
	WRITE_REG(RDFLD, (UInt32)result);

	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_sub_flags(emu, result);

	EMULATOR_STUB(rsc_1s,inst);
}

void impl_rsc_2(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, shifter_operand2(emu, inst) - RN - (1-C_FLAG));

	EMULATOR_STUB(rsc_2,inst);
}

void impl_rsc_2s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_sub64_w_carry(emu, shifter_operand2(emu, inst), RN, 1-C_FLAG);
	WRITE_REG(RDFLD, (UInt32)result);

	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_sub_flags(emu, result);

	EMULATOR_STUB(rsc_2s,inst);
}

void impl_rsc_3(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, shifter_operand3(emu, inst) - RN - (1-C_FLAG));

	EMULATOR_STUB(rsc_3,inst);
}

void impl_rsc_3s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_sub64_w_carry(emu, shifter_operand3(emu, inst), RN, 1-C_FLAG);
	WRITE_REG(RDFLD, (UInt32)result);

	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else
		set_sub_flags(emu, result);

	EMULATOR_STUB(rsc_3s,inst);
}

/*AND*/
void impl_and_1(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN & shifter_operand1(emu, inst));

	EMULATOR_STUB(and_1,inst);
}

void impl_and_1s(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN & shifter_operand1_s(emu, inst));
	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_move_flags(emu, RD);

	EMULATOR_STUB(and_1s,inst);
}

void impl_and_2(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN & shifter_operand2(emu, inst));

	EMULATOR_STUB(and_2,inst);
}

void impl_and_2s(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN & shifter_operand2_s(emu, inst));
	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_move_flags(emu, RD);

	EMULATOR_STUB(and_2s,inst);
}

void impl_and_3(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN & shifter_operand3(emu, inst));

	EMULATOR_STUB(and_3,inst);
}

void impl_and_3s(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN & shifter_operand3_s(emu, inst));
	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_move_flags(emu, RD);

	EMULATOR_STUB(and_3s,inst);
}

/*EOR*/
void impl_eor_1(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN ^ shifter_operand1(emu, inst));

	EMULATOR_STUB(eor_1,inst);
}

void impl_eor_1s(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN ^ shifter_operand1_s(emu, inst));
	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_move_flags(emu, RD);

	EMULATOR_STUB(eor_1s,inst);
}

void impl_eor_2(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN ^ shifter_operand2(emu, inst));

	EMULATOR_STUB(eor_2,inst);
}

void impl_eor_2s(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN ^ shifter_operand2_s(emu, inst));
	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_move_flags(emu, RD);

	EMULATOR_STUB(eor_2s,inst);
}

void impl_eor_3(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN ^ shifter_operand3(emu, inst));

	EMULATOR_STUB(eor_3,inst);
}

void impl_eor_3s(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN ^ shifter_operand3_s(emu, inst));
	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_move_flags(emu, RD);

	EMULATOR_STUB(eor_3s,inst);
}

/*ORR*/
void impl_orr_1(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN | shifter_operand1(emu, inst));

	EMULATOR_STUB(orr_1,inst);
}

void impl_orr_1s(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN | shifter_operand1_s(emu, inst));
	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_move_flags(emu, RD);

	EMULATOR_STUB(orr_1s,inst);
}

void impl_orr_2(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN | shifter_operand2(emu, inst));

	EMULATOR_STUB(orr_2,inst);
}

void impl_orr_2s(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN | shifter_operand2_s(emu, inst));
	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_move_flags(emu, RD);

	EMULATOR_STUB(orr_2s,inst);
}

void impl_orr_3(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN | shifter_operand3(emu, inst));

	EMULATOR_STUB(orr_3,inst);
}

void impl_orr_3s(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN | shifter_operand3_s(emu, inst));
	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_move_flags(emu, RD);

	EMULATOR_STUB(orr_3s,inst);
}

/*BIC*/
void impl_bic_1(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN & ~shifter_operand1(emu, inst));

	EMULATOR_STUB(bic_1,inst);
}

void impl_bic_1s(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN & ~shifter_operand1_s(emu, inst));
	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_move_flags(emu, RD);

	EMULATOR_STUB(bic_1s,inst);
}

void impl_bic_2(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN & ~shifter_operand2(emu, inst));

	EMULATOR_STUB(bic_2,inst);
}

void impl_bic_2s(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN & ~shifter_operand2_s(emu, inst));
	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_move_flags(emu, RD);

	EMULATOR_STUB(bic_2s,inst);
}

void impl_bic_3(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN & ~shifter_operand3(emu, inst));

	EMULATOR_STUB(bic_3,inst);
}

void impl_bic_3s(armulator *emu, arm_inst_t inst)
{
	WRITE_REG(RDFLD, RN & ~shifter_operand3_s(emu, inst));
	if (RDFLD==15)
		WRITE_CPSR(SPSR);
	else 
		set_move_flags(emu, RD);

	EMULATOR_STUB(bic_3s,inst);
}

/*CMP*/
void impl_cmp_1s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_sub64(emu, RN, shifter_operand1(emu, inst));
	set_sub_flags(emu, result);

	EMULATOR_STUB(cmp_1s,inst);
}

void impl_cmp_2s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_sub64(emu, RN, shifter_operand2(emu, inst));
	set_sub_flags(emu, result);

	EMULATOR_STUB(cmp_2s,inst);
}

void impl_cmp_3s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_sub64(emu, RN, shifter_operand3(emu, inst));
	set_sub_flags(emu, result);

	EMULATOR_STUB(cmp_3s,inst);
}

/*CMN*/
void impl_cmn_1s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_add64(emu, RN, shifter_operand1(emu, inst));
	set_add_flags(emu, result);

	EMULATOR_STUB(cmn_1s,inst);
}

void impl_cmn_2s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_add64(emu, RN, shifter_operand2(emu, inst));
	set_add_flags(emu, result);

	EMULATOR_STUB(cmn_2s,inst);
}

void impl_cmn_3s(armulator *emu, arm_inst_t inst)
{
	UInt64 result;
	result = impl_add64(emu, RN, shifter_operand3(emu, inst));
	set_add_flags(emu, result);

	EMULATOR_STUB(cmn_3s,inst);
}

/*TST*/
void impl_tst_1s(armulator *emu, arm_inst_t inst)
{
	UInt32 result;
	result = RN & shifter_operand1_s(emu, inst);
	set_move_flags(emu, result);

	EMULATOR_STUB(tst_1s,inst);
}

void impl_tst_2s(armulator *emu, arm_inst_t inst)
{
	UInt32 result;
	result = RN & shifter_operand2_s(emu, inst);
	set_move_flags(emu, result);

	EMULATOR_STUB(tst_2s,inst);
}

void impl_tst_3s(armulator *emu, arm_inst_t inst)
{
	UInt32 result;
	result = RN & shifter_operand3_s(emu, inst);
	set_move_flags(emu, result);

	EMULATOR_STUB(tst_3s,inst);
}

/*TEQ*/
void impl_teq_1s(armulator *emu, arm_inst_t inst)
{
	UInt32 result;
	result = RN ^ shifter_operand1_s(emu, inst);
	set_move_flags(emu, result);

	EMULATOR_STUB(teq_1s,inst);
}

void impl_teq_2s(armulator *emu, arm_inst_t inst)
{
	UInt32 result;
	result = RN ^ shifter_operand2_s(emu, inst);
	set_move_flags(emu, result);

	EMULATOR_STUB(teq_2s,inst);
}

void impl_teq_3s(armulator *emu, arm_inst_t inst)
{
	UInt32 result;
	result = RN ^ shifter_operand3_s(emu, inst);
	set_move_flags(emu, result);

	EMULATOR_STUB(teq_3s,inst);
}

static UInt32 disasm_shifter_operand1(arm_inst_t inst, char *buf)
{
	UInt32 imm = inst & 0xff;
	UInt32 shift = (inst>>7) & 0x1e;
	UInt32 result;

	result = rotate_right(imm, shift);
	return sprintf(buf, "#%d", result);
}

static UInt32 disasm_shifter_operand2(arm_inst_t inst, char *buf)
{
	UInt32 shift_imm = (inst>>7) & 0x01f;
	UInt8 type = (inst>>5) & 0x03;

	if (shift_imm==0 && type==3)
		return sprintf(buf, "%s, rrx", arm_regnames[RMFLD]);
	else if (shift_imm==0 && type==0)
		return sprintf(buf, "%s", arm_regnames[RMFLD]);
	else
		return sprintf(buf, "%s, %s #%d",
			arm_regnames[RMFLD], arm_shift[type], shift_imm);
}

static UInt32 disasm_shifter_operand3(arm_inst_t inst, char *buf)
{
	UInt8 type = (inst>>5) & 0x03;

	return sprintf(buf, "%s, %s %s",
		arm_regnames[RMFLD], arm_shift[type], arm_regnames[RSFLD]);
}

char *disasm_unop_1(arm_inst_t inst, arm_addr_t addr, char *buf)
{
	buf += sprintf(buf, "%s%s%s %s, ",
		arm_dpinames[DOPC], arm_conditional[COND], SFLD?"s":"",
		arm_regnames[RDFLD]);
	buf += disasm_shifter_operand1(inst, buf);
	buf += sprintf(buf, ";\n");
	return buf;
} 

char *disasm_unop_2(arm_inst_t inst, arm_addr_t addr, char *buf)
{
	buf += sprintf(buf, "%s%s%s %s, ",
		arm_dpinames[DOPC], arm_conditional[COND], SFLD?"s":"",
		arm_regnames[RDFLD]);
	buf += disasm_shifter_operand2(inst, buf);
	buf += sprintf(buf, ";\n");
	return buf;
}

char *disasm_unop_3(arm_inst_t inst, arm_addr_t addr, char *buf)
{
	buf += sprintf(buf, "%s%s%s %s, ",
		arm_dpinames[DOPC], arm_conditional[COND], SFLD?"s":"",
		arm_regnames[RDFLD]);
	buf += disasm_shifter_operand3(inst, buf);
	buf += sprintf(buf, ";\n");
	return buf;
}

char *disasm_binop_1(arm_inst_t inst, arm_addr_t addr, char *buf)
{
	buf += sprintf(buf, "%s%s%s %s, %s, ",
		arm_dpinames[DOPC], arm_conditional[COND], SFLD?"s":"",
		arm_regnames[RDFLD], arm_regnames[RNFLD]);
	buf += disasm_shifter_operand1(inst, buf);
	buf += sprintf(buf, ";\n");
	return buf;
}

char *disasm_binop_2(arm_inst_t inst, arm_addr_t addr, char *buf)
{
	buf += sprintf(buf, "%s%s%s %s, %s, ",
		arm_dpinames[DOPC], arm_conditional[COND], SFLD?"s":"",
		arm_regnames[RDFLD], arm_regnames[RNFLD]);
	buf += disasm_shifter_operand2(inst, buf);
	buf += sprintf(buf, ";\n");
	return buf;
}

char *disasm_binop_3(arm_inst_t inst, arm_addr_t addr, char *buf)
{
	buf += sprintf(buf, "%s%s%s %s, %s, ",
		arm_dpinames[DOPC], arm_conditional[COND], SFLD?"s":"",
		arm_regnames[RDFLD], arm_regnames[RNFLD]);
	buf += disasm_shifter_operand3(inst, buf);
	buf += sprintf(buf, ";\n");
	return buf;
}

char *disasm_tst_1(arm_inst_t inst, arm_addr_t addr, char *buf)
{
	buf += sprintf(buf, "%s%s %s, ",
		arm_dpinames[DOPC], arm_conditional[COND],
		arm_regnames[RNFLD]);
	buf += disasm_shifter_operand1(inst, buf);
	buf += sprintf(buf, ";\n");
	return buf;
}

char *disasm_tst_2(arm_inst_t inst, arm_addr_t addr, char *buf)
{
	buf += sprintf(buf, "%s%s %s, ",
		arm_dpinames[DOPC], arm_conditional[COND],
		arm_regnames[RNFLD]);
	buf += disasm_shifter_operand2(inst, buf);
	buf += sprintf(buf, ";\n");
	return buf;
}

char *disasm_tst_3(arm_inst_t inst, arm_addr_t addr, char *buf)
{
	buf += sprintf(buf, "%s%s %s, ",
		arm_dpinames[DOPC], arm_conditional[COND],
		arm_regnames[RNFLD]);
	buf += disasm_shifter_operand3(inst, buf);
	buf += sprintf(buf, ";\n");
	return buf;
}
