/***************************************************************************
                          emumem.h  -  description
                             -------------------
    begin                : Wed Sep 26 2001
    copyright            : (C) 2001 CEA and Universit Paris XI Orsay
    author               : Gilles Mouchard
    email                : gilles.mouchard@lri.fr, gilles.mouchard@cea.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef __EMU_MEMORY_H__
#define __EMU_MEMORY_H__

#include "misc.h"

namespace emulator {

const unsigned int MEMORY_PAGE_SIZE = 4096;
const unsigned int PRIMARY_MEMORY_HASH_TABLE_SIZE = 4096;
const unsigned int SECONDARY_MEMORY_HASH_TABLE_SIZE = 16;

typedef struct memory_page_table_entry_t
{
	target_addr_t addr;
	struct memory_page_table_entry_t *next;
	byte_t *storage;
} memory_page_table_entry_t;

typedef struct
{
	memory_page_table_entry_t *pte[SECONDARY_MEMORY_HASH_TABLE_SIZE];
} secondary_memory_hash_table_t;

class memory
{

	public:

		/* constructor */
		memory();

		/* destructor */
		~memory();

		/* copy constructor, not implemented */
		memory(const memory&);

		/* free all memory pages */
		void reset();

		/* read operations */
		byte_t read_byte(target_addr_t addr);
		halfword_t read_half_word(target_addr_t addr);
		word_t read_word(target_addr_t addr);
		dword_t read_dword(target_addr_t addr);

		void read_block(void *buf, target_addr_t addr, unsigned int size);

		/* write operations */
		void write_byte(target_addr_t addr, byte_t value);
		void write_half_word(target_addr_t addr, halfword_t value);
		void write_word(target_addr_t addr, word_t value);
		void write_dword(target_addr_t addr, dword_t value);

		void set_block(target_addr_t addr, byte_t value, unsigned int size);
		void write_block(target_addr_t, void *buf, unsigned int size);

		/* some usefule utilities */
		target_addr_t align_to_page_boundary(target_addr_t addr);

		/* memory test , true->succeed */
		//bool test();

		/* if the fist read is to address 0, this will crash, so beware */
		word_t read_word_fast(target_addr_t addr) {

 			memory_page_table_entry_t *pte; 
			target_addr_t offset = addr % MEMORY_PAGE_SIZE; 

		    if (cached_addr==addr-offset) pte = cached_pte; 
		    else {
				cached_pte = pte = get_page(addr);
				cached_addr = addr-offset;
			}

#if WORDS_BIGENDIAN==TARGET_LITTLE_ENDIAN
		    return swap_word(*((word_t *)(pte->storage+offset)));
#else
		    return *((word_t *)(pte->storage+offset));
#endif
		}                               

		/* return the number of pages allocated */
		unsigned int get_page_count() {
			return page_count;
		}
			   
	private:

		static halfword_t swap_half_word(halfword_t val) {
			return (val>>8) | (val<<8);
		}

		static word_t swap_word(word_t val) {
			return (val>>24) | ((val>>8)&0xFF00) |
				   ((val&0xFF00)<<8) |  (val<<24) ;
		}

		static dword_t swap_dword(dword_t val) {
			return ((dword_t)swap_word((word_t)val)<<32) |
				    (dword_t)swap_word((word_t)(val>>32));
		}

		static UInt32 hash1(target_addr_t addr)
		{
			return (addr / MEMORY_PAGE_SIZE) % PRIMARY_MEMORY_HASH_TABLE_SIZE;
		}

		static UInt32 hash2(target_addr_t addr)
		{
			return (addr / MEMORY_PAGE_SIZE / PRIMARY_MEMORY_HASH_TABLE_SIZE) %
			   SECONDARY_MEMORY_HASH_TABLE_SIZE;
		}

 		memory_page_table_entry_t *get_page(target_addr_t);
 		memory_page_table_entry_t *search_page(target_addr_t);

		secondary_memory_hash_table_t 
			*primary_hash_table[PRIMARY_MEMORY_HASH_TABLE_SIZE];

		unsigned int page_count;	/*stats information */

		/* for possibly sequential reads by read_word_fast */
		target_addr_t cached_addr;
		memory_page_table_entry_t *cached_pte;

};

} /* namespace */

#endif
