/***************************************************************************
                          memory.h  -  description
                             -------------------
    begin                : Wed Sep 26 2001
    copyright            : (C) 2001 CEA and Universit Paris XI Orsay
    author               : Gilles Mouchard
    email                : gilles.mouchard@lri.fr, gilles.mouchard@cea.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef __MEMORY_H__
#define __MEMORY_H__

#include "misc.h"
#include <stdio.h>

#define MEMORY_PAGE_SIZE 4096
#define PRIMARY_MEMORY_HASH_TABLE_SIZE 4096
#define SECONDARY_MEMORY_HASH_TABLE_SIZE 16

typedef struct
{
	/* Memory interface */
	byte_t (*mem_read_byte)(void *, target_addr_t);
	void (*mem_write_byte)(void *, target_addr_t, byte_t);
	halfword_t (*mem_read_half_word)(void *, target_addr_t);
	void (*mem_write_half_word)(void *, target_addr_t, halfword_t);
	word_t (*mem_read_word)(void *, target_addr_t);
	void (*mem_write_word)(void *, target_addr_t, word_t);
	dword_t (*mem_read_dword)(void *, target_addr_t);
	void (*mem_write_dword)(void *, target_addr_t, dword_t);
	void (*mem_set)(void *, target_addr_t, byte_t, int);
	void (*mem_read)(void *, void *, target_addr_t, int);
	void (*mem_write)(void *, target_addr_t, void *, int);

	
	void *instance;
} memory_cb_t;

typedef struct memory_page_table_entry_t
{
	target_addr_t addr;
	struct memory_page_table_entry_t *next;
	byte_t *storage;
} memory_page_table_entry_t;

typedef struct
{
	memory_page_table_entry_t *pte[SECONDARY_MEMORY_HASH_TABLE_SIZE];
} secondary_memory_hash_table_t;

typedef struct
{
	secondary_memory_hash_table_t *primary_hash_table[PRIMARY_MEMORY_HASH_TABLE_SIZE];
	unsigned int page_count;	/*stats information */
} memory_t;

extern memory_t *mem;

#if defined(__cplusplus)
extern "C" {
#endif
extern memory_t *mem_create();
extern byte_t mem_read_byte(memory_t *mem, target_addr_t addr);
extern void mem_write_byte(memory_t *mem, target_addr_t addr, byte_t value);
extern halfword_t mem_read_half_word(memory_t *mem, target_addr_t addr);
extern void mem_write_half_word(memory_t *mem, target_addr_t addr, halfword_t value);
extern word_t mem_read_word(memory_t *mem, target_addr_t addr);
extern void mem_write_word(memory_t *mem, target_addr_t addr, word_t value);
extern dword_t mem_read_dword(memory_t *mem, target_addr_t addr);
extern void mem_write_dword(memory_t *mem, target_addr_t addr, dword_t value);
extern halfword_t mem_read_half_word_byte_reverse(memory_t *mem, target_addr_t addr);
extern void mem_write_half_word_byte_reverse(memory_t *mem, target_addr_t addr, halfword_t value);
extern word_t mem_read_word_byte_reverse(memory_t *mem, target_addr_t addr);
extern void mem_write_word_byte_reverse(memory_t *mem, target_addr_t addr, word_t value);
extern dword_t mem_read_dword_byte_reverse(memory_t *mem, target_addr_t addr);
extern void mem_write_dword_byte_reverse(memory_t *mem, target_addr_t addr, dword_t value);
extern void mem_set(memory_t *mem, target_addr_t addr, byte_t value, int size);
extern void mem_read(memory_t *mem, void *buf, target_addr_t addr, int size);
extern void mem_write(memory_t *mem, target_addr_t, void *buf, int size);
extern target_addr_t mem_align_to_page_boundary(target_addr_t addr);
extern void mem_delete(memory_t *mem);
extern void mem_save(memory_t *mem, FILE *f);
extern void mem_restore(memory_t *mem, FILE *f);
extern void mem_duplicate(memory_t *mem, memory_cb_t *cb);
extern  memory_page_table_entry_t *mem_get_page(memory_t *, target_addr_t);
#if defined(__cplusplus)
}
#endif

/*for little endian target, reverse byte order while accessing memory*/
#ifdef TARGET_LITTLE_ENDIAN
#define MEM_READ_BYTE(addr) mem_read_byte(mem, addr)
#define MEM_WRITE_BYTE(addr, value) mem_write_byte(mem, addr, value)
#define MEM_READ_HALF_WORD(addr) mem_read_half_word_byte_reverse(mem, addr)
#define MEM_WRITE_HALF_WORD(addr, value) mem_write_half_word_byte_reverse(mem, addr, value)
#define MEM_READ_WORD(addr) mem_read_word_byte_reverse(mem, addr)
#define MEM_WRITE_WORD(addr, value) mem_write_word_byte_reverse(mem, addr, value)
#define MEM_READ_DWORD(addr) mem_read_dword_byte_reverse(mem, addr)
#define MEM_WRITE_DWORD(addr, value) mem_write_dword_byte_reverse(mem, addr, value)
#else 
#define MEM_READ_BYTE(addr) mem_read_byte(mem, addr)
#define MEM_WRITE_BYTE(addr, value) mem_write_byte(mem, addr, value)
#define MEM_READ_HALF_WORD(addr) mem_read_half_word(mem, addr)
#define MEM_WRITE_HALF_WORD(addr, value) mem_write_half_word(mem, addr, value)
#define MEM_READ_WORD(addr) mem_read_word(mem, addr)
#define MEM_WRITE_WORD(addr, value) mem_write_word(mem, addr, value)
#define MEM_READ_DWORD(addr) mem_read_dword(mem, addr)
#define MEM_WRITE_DWORD(addr, value) mem_write_dword(mem, addr, value)
#endif

#define MEM_SET(addr, value, size) mem_set(mem, addr, value, size)
#define MEM_READ(buf, addr, size) mem_read(mem, buf, addr, size)
#define MEM_WRITE(addr, buf, size) mem_write(mem, addr, buf, size)

#if TARGET_LITTLE_ENDIAN==WORDS_BIGENDIAN
#define SWAP_WORD(val) ((val>>24)|((val>>8)&0xFF00)|((val&0xFF00)<<8)|(val<<24))
#else
#define SWAP_WORD(val) val 
#endif

#ifdef NO_PAGE_CACHE

#define MEM_READ_WORD_FAST(ret, addr, last_addr, last_pte, type) {\
	ret = MEM_READ_WORD(addr);}

#define MEM_WRITE_WORD_FAST(ret, addr, last_addr, last_pte, type) {\
    MEM_WRITE_WORD(addr, ret);}

#define MEM_READ_WORD_FAST_NO_UPDATE(ret, addr, last_addr, last_pte, type) {\
	ret = MEM_READ_WORD(addr);}

#define MEM_WRITE_WORD_FAST_NO_UPDATE(ret, addr, last_addr, last_pte, type) {\
    MEM_WRITE_WORD(addr, ret);}

#define MEM_READ_WORD_FAST2_NO_UPDATE(ret, addr, last_addr, last_pte, last_addr2,last_pte2, type) {\
	ret = MEM_READ_WORD(addr);}

#else

#define MEM_READ_WORD_FAST(ret, addr, last_addr, last_pte, type) {\
    memory_page_table_entry_t *pte; \
    target_addr_t offset = addr % MEMORY_PAGE_SIZE; \
    if (last_addr==addr-offset) pte = last_pte; \
    else last_pte=pte=mem_get_page(mem, addr), last_addr = addr-offset; \
    ret = *((type *)(pte->storage+offset));\
	ret = SWAP_WORD(ret);}

#define MEM_WRITE_WORD_FAST(ret, addr, last_addr, last_pte, type) {\
    memory_page_table_entry_t *pte; \
    target_addr_t offset = addr % MEMORY_PAGE_SIZE; \
    if (last_addr==addr-offset) pte = last_pte; \
    else last_pte=pte=mem_get_page(mem, addr), last_addr = addr-offset; \
    *((type *)(pte->storage+offset))=SWAP_WORD(val);}

#define MEM_READ_WORD_FAST_NO_UPDATE(ret, addr, last_addr, last_pte, type) {\
    memory_page_table_entry_t *pte; \
    target_addr_t offset = addr % MEMORY_PAGE_SIZE; \
    if (last_addr==addr-offset) pte = last_pte; \
    else pte=mem_get_page(mem, addr); \
    ret = *((type *)(pte->storage+offset)); \
	ret = SWAP_WORD(ret);}

#define MEM_WRITE_WORD_FAST_NO_UPDATE(ret, addr, last_addr, last_pte, type) {\
    memory_page_table_entry_t *pte; \
    target_addr_t offset = addr % MEMORY_PAGE_SIZE; \
    if (last_addr==addr-offset) pte = last_pte; \
    else pte=mem_get_page(mem, addr); \
    *((type *)(pte->storage+offset))=SWAP_WORD(ret);}

#define MEM_READ_WORD_FAST2_NO_UPDATE(ret, addr, last_addr, last_pte, last_addr2,last_pte2, type) {\
    memory_page_table_entry_t *pte; \
    target_addr_t offset = addr % MEMORY_PAGE_SIZE; \
    if (last_addr==addr-offset) pte = last_pte; \
    else if (last_addr2==addr-offset) pte = last_pte2; \
    else pte=mem_get_page(mem, addr); \
    ret = *((type *)(pte->storage+offset)); \
	ret = SWAP_WORD(ret);}

#endif

#endif
