
#ifndef __ITINERARY_H__
#define __ITINERARY_H__

#include <vector>
#include <string>
#include "config.h"
#include "bittypes.h"
#include "token.h"
#include "manager.h"
#include <stdio.h>
#include <assert.h>

namespace osm {
typedef class object Object;

/** Action to perform. */
typedef class action {

	public:

		/** Enum for actions. */
		/** Discard is unconditional release, all discards can be
			replaced by release. I used it since it is 10% faster by saving
			the useless test (GoodToReclaim) which is virtual and 
			cannot be inlined */
		typedef enum {
			ACTION_INQUIRE,
			ACTION_ALLOCATE,
			ACTION_RELEASE,
			ACTION_DISCARD,
		} ACTION_TYPE;

		/** constructor */
		action(ACTION_TYPE type, Manager *boss, Token **token, uint32_t id) :
			type(type), boss(boss), token(token), id(id) {}

		/** destructor */
		~action() {}

		/** set the id of an action */
		void setID(uint32_t id) {this->id = id;}

		/** try to fire the action, return true if firable */
		bool tryFire(Object *obj) {
			switch (type) {
				case ACTION_INQUIRE:
					return boss->isTokenFreeForInquire(id, obj);
				case ACTION_ALLOCATE:
					return boss->isTokenFreeForAlloc(id, obj);
				case ACTION_RELEASE:
					return boss->isTokenGoodToReclaim(*token, obj);
				default:
					return true;
			}
		}

		/** fire the action */
		void fire(Object *obj) {
#ifdef COUNT_ACTIONS
			action_fired++;
#endif
			switch(type) {
				case ACTION_INQUIRE:
					*token = boss->inquireToken(id, obj);
					break;
				case ACTION_ALLOCATE:
					*token = boss->allocateToken(id, obj);
					break;
				case ACTION_RELEASE:
					if (boss->reclaimToken(*token, obj)) *token = NULL;
					break;
				case ACTION_DISCARD:
					boss->discardToken(*token, obj); *token = NULL;
					break;
				default:
					break;
			}
		}

	private:

		ACTION_TYPE type;
		Manager *boss;			/** the token manager to turn to*/
		Token 	**token;		/** the place to hold grabbed token*/
		uint32_t id;

} Action;

typedef class inode Node;
typedef class iedge Edge;

class iedge {

	public:

		/** constructor */
		iedge(Node *src, Node *dst) : src(src), dst(dst), actions() {}

		/** destructor */
		~iedge() {
			std::vector<Action *>::iterator it;
			for (it=actions.begin();it!=actions.end();it++)
				delete *it;
		}

		/** try to fire the actions associated with an edge, return true
		 *  if all are ready to fire. */
		bool tryFire(Object *obj) {
			std::vector<Action *>::iterator it;
			for (it=actions.begin();it!=actions.end();it++) {
				if (!(*it)->tryFire(obj)) return false;
			}
			return true;
		}

		/** fire the actions associated with an edge */
		Node *fire(Object *obj) {
			std::vector<Action *>::iterator it;
			for (it=actions.begin();it!=actions.end();it++)
				(*it)->fire(obj);
			return dst;
		}

		/** append a new action */
		void addAction(Action *action) {actions.push_back(action);}

		/** set action ID */
		void setActionID(int i, uint32_t id) {actions[i]->setID(id);}

	private:

		Node *src, *dst;
		std::vector<Action *> actions;
};

/** A stop in an itinerary. */
class inode {

	public:

		/** constructor */
		inode(const std::string& name) : name(name), outEdges() {}

		/** destructor */
		~inode() {}

		/** Check if all actions on an edge are firable */
		Edge *validEdge(Object *obj) {
			std::vector<Edge *>::iterator it;
			for (it=outEdges.begin();it!=outEdges.end();it++)
				if ((*it)->tryFire(obj)) return *it;
			return NULL;
		}

		/** move onto the next node */
		Node *moveOn(Object *obj) {
			Edge *eg = validEdge(obj);
			return eg?eg->fire(obj):this;
		}

		/** get id */
		const std::string& getName() {return name;}

		/** append a new outedge */
		void addEdge(Edge *edge) {outEdges.push_back(edge);}

	private:

		std::string name;
		std::vector<Edge *> outEdges;
};

/** an abstract itinerary of an operation */
typedef class itinerary {

	public:

		/** constructor */
		itinerary(int t) : type(t), currentNode(NULL),
			nodes(), edges(), tpbuffer() {}

		/** destructor */
		virtual ~itinerary();

		/** activate the itinerary */
		bool activate(Object *obj) {
			assert(currentNode);
			currentNode = currentNode->moveOn(obj);
			return currentNode==NULL;
		}

		/** reset for reuse */
		void reset() {currentNode = *nodes.begin();}

		/** cleanup by discarding all tokens */
		void cleanup(Object *obj) {                             
			std::vector<Token *>::iterator it;
			for (it = tpbuffer.begin(); it != tpbuffer.end(); it++) {
				if ((*it)!=NULL) {(*it)->discard(obj); *it=NULL;}
			}
		}

		/** return the type ID of the itinerary */
		int getType() {return type;}

		/** return the current node of the itinerary */
		Node *where() {return currentNode;}

		/** create a new Node */
		Node* newNode(const std::string &name) {
			Node *node = new Node(name);
			nodes.push_back(node);
			return node;
		}

		/** create a new edge */
		Edge* newEdge(Node *src, Node *dst) {
			Edge *edge = new Edge(src, dst);
			edges.push_back(edge);
			src->addEdge(edge);
			return edge;
		}

		/** make this an abstract class */
		virtual void setIds(uint32_t ids[]) = 0;

	protected:

		int type;
		Node *currentNode;
		std::vector<Node *> nodes;
		std::vector<Edge *> edges;
		std::vector<Token *> tpbuffer;

	friend class object;

} Itinerary;

}

#endif
