
#ifndef __ARMSIM_H__
#define __ARMSIM_H__

#include "armemul.h"
#include "simple_sched.h"
#include "misc.h"
#include "itinerary_pool.h"
#include "derived_itinerary.h"
#include "icache.h"
#include "dcache.h"
#include "itlb.h"
#include "dtlb.h"
#include <string>
#include <map>

typedef class osm::object Object;
typedef class osm::token Token;
typedef class osm::inode Node;
typedef class osm::iedge Edge;
typedef class osm::action Action;
typedef class osm::itinerary Itinerary;
typedef class osm::scheduler Scheduler;


namespace simulator {

	/*maximum number of live instruction in the pipeline */
	const unsigned int MAX_INSTS = 6;

	/*maximum number of u-ops in pipeline*/
	const unsigned int MAX_UOPS = 32;

	class arm_simulator {

		public:

			/* constructor */
			arm_simulator(bool verbose);

			/* destructor */
			~arm_simulator();

			void reset();

			void reset_stats();
			void dump_stats();

			bool clock_tick();
			void run(UInt64);

			void load_program(const char *,int argc,char *argv[],char *env[]);
			void load_fpe(const char *);

			itn_pool *itnq;

/* the buffer holding the uops(micro-op) decoded from a CISC style instruction
 *    CISC style instr. in ARM includes ldm, stm, bl*/
			std::vector<Itinerary *> uopBuffer;
			int uopBufferSize;

/* the CISC instruction corresponding to uopBuffer*/
			arm_inst_t uopInst;
			arm_addr_t uopAddr;

			/* current mode */
			bool in_speculation;

			/* emulator */
			emulator::arm_emulator emu;

			/* interface to scheduler */
			void updateOnClockTick();
			bool decodeAndExecuteObject(Object *obj, Token *);

			void retireObject(Object *obj) {
				itnq->release((obj)->getItinerary());
			}

			bool canAwake() {
				return fetchManager->isTokenFreeForAlloc(0, NULL);
			}

			void awakeObject(Object *obj) {
				fetchManager->allocateToken(0, obj);
			}

			void branchResolved(Object *);

			/* added by Xinping */
			void ns_enqueue(unsigned address, int x);
			emulator::ns_data ns_dequeue();
			int ns_input_queue_size(int x);
			int ns_output_queue_size();
			void setAddress(unsigned int x);

		private:
			UInt64 uop_count; /*total committed uops*/
			UInt64 null_count; /*total instruction with predicate false*/
			UInt64 branch_count;  /*total branchs executed*/

			/*the fetching stage manager*/
			Manager *fetchManager;
			Manager *bufferManager;
			Manager *exeManager;
			std::map<std::string, Manager *> managers;
			
			Scheduler *sched;
			UInt64 cycleNum;

			BIU biu;
			ICache icache;
			DCache dcache;
			ITLB itlb;
			DTLB dtlb;

			void init_itineraries();
			void init_managers();

			void delete_itineraries();
			void delete_managers();

			void dump_instruction_counters();

	};
}

#define initObject(obj, inst, addr) {\
	obj->init(inst, addr); \
}

#define squashObject(obj) { \
	sched->squash(obj->getDynId()); \
}

#endif
