
#include "armsim.h"
#include "decode.h"
#include "decode_itin.h"

#if HAVE_SYS_RESOURCE_H && HAVE_SYS_TIME_H
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#endif

#include <vector>
#include "syscall.h"
#include "dpi.h"
#include "branch.h"
#include "mult.h"
#include "ldst.h"
#include "statreg.h"
#include "nwfpe.h"
#include "derived_itinerary.h"
#include "derived_managers.h"

#include "auto_itin.h"
#include "arm_dec.h"

using namespace simulator;
using std::vector;
using std::map;
using std::string;

void arm_simulator::init_itineraries()
{
	unsigned int i;
	std::vector<Itinerary *> itns;

#define xxx(a,b,c) \
	{for (i=0;i<c;i++) itns.push_back(new a##Itinerary(managers));}
#include "itin_list.h"

	itnq = new itn_pool(itns);
}

void arm_simulator::init_managers()
{
	managers["IF"] = fetchManager = 
		new FetchManager("fetch", this, icache, itlb);
	managers["ID"] = new DecodeManager("decode");
	managers["EX"] = exeManager = new LatencyManager("execute");
	managers["BF"] = bufferManager = new BFManager("buffer", dcache, dtlb);
	managers["WB"] = new Manager(1, "writeback");
	managers["ME"] = new Manager(1, "mulcontrol");
	managers["RF"] = new RegFileManager(16, "regfile", this);
}

void arm_simulator::delete_itineraries()
{
	delete itnq;
}

void arm_simulator::delete_managers()
{
	map<string, Manager *>::iterator it;
	for (it=managers.begin(); it!=managers.end(); it++)
		delete it->second;
}

arm_simulator::arm_simulator(bool verbose) :
	uopBuffer(), uopBufferSize(0), in_speculation(false),
	emu(verbose, true),
   	biu("biu"), icache("icache", biu), dcache("dcache", biu), 
	itlb("itlb", biu), dtlb("dtlb", biu)
{
	uopBuffer.resize(16);
	init_managers();
	init_itineraries();
	sched = new Scheduler(MAX_UOPS+MAX_INSTS, this);
}

arm_simulator::~arm_simulator()
{
	delete sched;
	delete_managers();
	delete_itineraries();
}

void arm_simulator::reset_instruction_counters()
{
	uop_count = 0;
	null_count = 0;
	branch_count = 0;
	emu.reset_instruction_counters();
}

void arm_simulator::setAddress(unsigned int x){
   emu.pe_address = x; 
}

void arm_simulator::ns_enqueue(unsigned address, int x){
   emu.input_queue[address].push(emulator::ns_data(address, x)); 
}

emulator::ns_data arm_simulator::ns_dequeue(){
  emulator::ns_data a = emu.output_queue.front();
  emu.output_queue.pop();
  return a;
}

int arm_simulator::ns_input_queue_size(int x){
  return emu.input_queue[x].size();
}

int arm_simulator::ns_output_queue_size(){
  return emu.output_queue.size();
}




void arm_simulator::load_program(const char *fname,
	int argc,char *argv[],char *env[])
{
	emu.load_program(fname, argc, argv, env);
	reset_stats();
	emu.set_running();
	static_cast<FetchManager*>(fetchManager)->setProgCounter(emu.get_pc());
}

void arm_simulator::load_fpe(const char *fname)
{
	emu.load_fpe(fname);
}

void arm_simulator::clock_tick()
{
	sched->tick();
	cycleNum++;
}
void arm_simulator::check_clock_tick()
{
	if ((cycleNum%1000)==0) fputs(".", stderr);
  if (emu.is_running()){
	sched->tick();
	cycleNum++;
  }
}


void arm_simulator::run(UInt64 max_count)
{

#if HAVE_SYS_RESOURCE_H && HAVE_SYS_TIME_H
	struct timeval begin_u, end_u, begin_s, end_s;			
	struct rusage usg;										
	float user_time, sys_time;								
	getrusage(RUSAGE_SELF, &usg);							 
	begin_u = usg.ru_utime;								   
	begin_s = usg.ru_stime;								   
#endif														

	while (emu.is_running() && emu.get_icount()<max_count) {
		clock_tick();
	}
	dump_stats();

#if HAVE_SYS_RESOURCE_H && HAVE_SYS_TIME_H
	getrusage(RUSAGE_SELF, &usg);							 
	end_u = usg.ru_utime;									 
	end_s = usg.ru_stime;									 
	user_time = (end_u.tv_sec+end_u.tv_usec/1000000.0)-	   
		(begin_u.tv_sec+begin_u.tv_usec/1000000.0);		   
	sys_time = (end_s.tv_sec+end_s.tv_usec/1000000.0)-		
		(begin_s.tv_sec+begin_s.tv_usec/1000000.0);		   
	fprintf(stderr,
		"Total user time  : %.3f sec.\n"		   
		"Total system time: %.3f sec.\n"					  
		"Simulation speed : %.3e inst/sec or %.3e cyc/sec.\n",				
		user_time, sys_time, emu.get_icount()/(user_time+sys_time),
		cycleNum/(user_time+sys_time));	
#endif														
}

void arm_simulator::reset_stats()
{
	reset_instruction_counters();						 
	cycleNum = 0;
}

void arm_simulator::dump_stats()
{
	/* print out stats information */
	dump_instruction_counters();						  
	fprintf(stderr, "IPC:   %.4f\n", (double)emu.get_icount()/cycleNum);

    icache.PrintStats();
    itlb.PrintStats();
    dcache.PrintStats();
    dtlb.PrintStats();
    biu.PrintStats();

	fprintf(stderr,
		"Total cycles     : ");
	dump_int64(cycleNum, stderr);		 
	fprintf(stderr,
		"\nEquivalent time on 206.4MHz host: %.4f sec.\n",
		(double)cycleNum/206438400.0);		 
															  
	fprintf(stderr, "Total memory pages allocated : %d\n",
		emu.mem.get_page_count());
}

#define dumpint64(i) \
	dump_int64(i, stderr); \
	if (i>1000000) { \
		fputs(" (", stderr);\
		dump_int64_smart(i, stderr);\
		fputs(")", stderr);\
	}

void arm_simulator::dump_instruction_counters()
{
	
	fputs("\n", stderr);

	fputs("Total fetched instructions :   ", stderr);
	dumpint64(static_cast<FetchManager*>(fetchManager)->getFetchCount());
	fputs("\n", stderr);

	fputs("Total committed instructions : ", stderr);
	dumpint64(emu.get_icount());
	fputs(" including ", stderr);
	dumpint64(null_count);
	fputs(" nullified\n", stderr);

	fputs("Total committed micro-ops :    ", stderr);
	dumpint64(uop_count);
	fputs("\n", stderr);
}


static const UInt16 cond_table[] =
{0xf0f0, 0x0f0f, 0xcccc, 0x3333,
 0xff00, 0x00ff, 0xaaaa, 0x5555,
 0x0c0c, 0xf3f3, 0xaa55, 0x55aa,
 0x0a05, 0xf5fa, 0xffff, 0x0000};

static inline UInt32 eval_cond(emulator::arm_emulator *emu, arm_inst_t inst)
{
    UInt32 val;
    val = (cond_table[COND] >> (CPSR>>28))&1;
	return val;
}


void arm_simulator::updateOnClockTick()
{
	/* benefit from inlining */
	static_cast<FetchManager*>(fetchManager)->updateOnClockTick();
	static_cast<BFManager*>(bufferManager)->updateOnClockTick();
	static_cast<LatencyManager*>(exeManager)->updateOnClockTick();

	icache.updateOnClockTick();
	dcache.updateOnClockTick();
	itlb.updateOnClockTick();
	dtlb.updateOnClockTick();
	biu.updateOnClockTick();
}

bool arm_simulator::decodeAndExecuteObject(Object *obj, Token *token)
{
	Itinerary *itn = NULL;

	/* see if more uops left */
	if (uopBufferSize==0) {

		arm_inst_t inst = uopInst = obj->getInst();
		uopAddr = obj->getAddr();

		if (!in_speculation) {

			/* Allocate the itinerary before executing the instruction
			   since the itinerary may use some register values and they
		 	  should be read before execution modifies the values */
			if (!eval_cond(&(emu), inst)) {
				/* condition false, use dummy */
				null_count++;
				itn = itnq->allocate(ITN_DUMMY);
				obj->nullify();
			}
			else itn = itn_main(this, inst);

#ifdef DEBUG
	arm_disassemble_to_stream(inst, obj->getAddr(), stderr);
#endif
			uop_count++;
			emu.execute(inst, obj->getAddr());
			if (emu.is_pc_updated()) in_speculation = true;
		}
		else itn = itn_main(this, inst);
		obj->setItinerary(itn, token);
	}
	else {
		Itinerary *itn = uopBuffer[--uopBufferSize];
		initObject(obj, uopInst, uopAddr);
		obj->setItinerary(itn, token);
		uop_count++;
	}
	return uopBufferSize>0;
}

void arm_simulator::branchResolved(Object *obj)
{
	branch_count++;
	static_cast<FetchManager*>(fetchManager)->setProgCounter(emu.get_pc());
	squashObject(obj);
	in_speculation = false;

	/* clean the uopbuffer */
	assert (uopBufferSize==0);
}

