/*************************************************************************
    Copyright (C) 2002,2003 Wei Qin
    See file COPYING for more information.

    This program is free software; you can redistribute it and/or modify    
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
*************************************************************************/

#include "misc.h"
#include "derived_managers.h"
#include "derived_itinerary.h"
#include "token.h"
#include "object.h"
#include "armsim.h"
#include "dcache.h"
#include <assert.h>

#ifdef DEBUG
#include <cstdio>
#endif

using namespace simulator;

bool FetchManager::isTokenGoodToReclaim(Token *token, Object *obj)
{
	/* if in cache line fill, wait till it finishes */
	return status==FETCH_DONE;
}

bool FetchManager::reclaimToken(Token *token, Object *obj)
{
	if (token->getObject()==obj) {
		token->setUnused();
		progCounter += sizeof(arm_inst_t);
	}
	return true;
}

Token *FetchManager::allocateToken(uint32_t id, Object *obj)
{
	if (moreToCome) {
 		moreToCome = sim->decodeAndExecuteObject(obj, tokens);
	}
	else {
		/* the first uop of a cisc or a risc*/
		arm_inst_t inst = sim->emu.fetch_inst(progCounter);
		initObject(obj, inst, progCounter);
 		moreToCome = sim->decodeAndExecuteObject(obj, tokens);
		status = FETCH_TLB;
		fetch_count++;
	}
	if (!moreToCome) tokens[0].setUsed(obj);
	return tokens;
}

void FetchManager::discardToken(Token *token, Object *obj)
{
	token->setUnused();
	status = FETCH_DONE;
}

void FetchManager::updateOnClockTick()
{
	if (status==FETCH_TLB) {
		if (itlb.read(progCounter, sizeof(arm_addr_t)))
			status = FETCH_CACHE;
	}
	if (status==FETCH_CACHE) {
		if (icache.read(progCounter, sizeof(arm_inst_t)))
			status = FETCH_DONE;
	}
	//fprintf(stderr, "address in fetch = %d \n", sim->emu.pe_address);
	assert(!moreToCome);
}

bool BFManager::isTokenGoodToReclaim(Token *token, Object *obj)
{
	/* allow early finish */
	return status==BF_DONE;
}

void BFManager::discardToken(Token *token, Object *obj)
{
	tokens->setUnused();
	status = BF_DONE;
}

static inline int isMemoryOp(Object *obj)
{
	int type = (obj->getItinerary())->getType();
	switch (type) {
		case ITN_LD1:
		case ITN_LD2:
			return 1;
		case ITN_LDM:
		case ITN_LDMLAST:
			return 2;
		case ITN_ST:
			return 3;
		case ITN_STM:
		case ITN_STMLAST:
			return 4;
		default:
			return 0;
	}
}

Token *BFManager::allocateToken(uint32_t id, Object *obj)
{
	int optype = isMemoryOp(obj);
	tokens[0].setUsed(obj);

	if (optype) {
		status = BF_TLB;
		isRead = optype==1 || optype==2;
		isMergeable = optype==2 || optype==4;
		addr = id;
	}
	return tokens;
}

void BFManager::updateOnClockTick()
{
	if (status==BF_TLB) {
		if (dtlb.read(addr, sizeof(arm_addr_t)))
			status = BF_CACHE;
	}
	if (status==BF_CACHE) {
		if ((isRead && dcache.read(addr, sizeof(arm_addr_t))) ||
			(!isRead && dcache.write(addr, 4, isMergeable)))
			status = BF_DONE;
	}
}

bool RegFileManager::isTokenFreeForAlloc(uint32_t id, Object *obj)
{
	/*-1 is dummy id for those with optional register operands*/
	/* PC = 15 is special, always free to grab */
	if (id==(uint32_t)-1) return true;

	/*allowing newer operation to preempt the token since WAW is nothing
	  to worry about in ARM*/
	return (tokens[id].isFree() ||
		obj->getDynId() > tokens[id].getObject()->getDynId());
}

bool RegFileManager::isTokenFreeForInquire(uint32_t id, Object *obj)
{
	return (id==(uint32_t)-1 || tokens[id].isFree());
}

Token *RegFileManager::allocateToken(uint32_t id, Object *obj)
{
	if (id==(uint32_t)-1) return NullToken;
	/*PC = 15 is always left free*/
	if (id!=15) tokens[id].setUsed(obj);
	return tokens+id;
}

bool RegFileManager::reclaimToken(Token *token, Object *obj)
{
	if (token==tokens+15 && obj->isActive())
		sim->branchResolved(obj);
	else if (obj == token->getObject())
		token->setUnused();
	return true;
}

void RegFileManager::discardToken(Token *token, Object *obj)
{
	if (obj == token->getObject())
		token->setUnused();
}
