/*************************************************************************
    Copyright (C) 2002,2003 Wei Qin
    See file COPYING for more information.

    This program is free software; you can redistribute it and/or modify    
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
*************************************************************************/

#ifndef __DERIVED_MANAGER__
#define __DERIVED_MANAGER__

#include "manager.h"
#include "token.h"
#include "object.h"
#include "icache.h"
#include "itlb.h"
#include "dcache.h"
#include "dtlb.h"
#include "simple_sched.h"
#include "armsim.h"
#include <string>

#include <cstdio>

namespace simulator {

typedef class fetch_manager : public Manager {

	public:

		/** constructor */
		/** the fetcher has two tokens, tokens[0] is for main operation
			in one instruction (the last uops), tokens[1] is for other uops*/
		fetch_manager(const std::string &name, arm_simulator *sim,
			ICache &icache, ITLB &itlb) : manager(1, name), 
			sim(sim), icache(icache), itlb(itlb), status(FETCH_DONE),
		   fetch_count(0), moreToCome(false) {}

		/** destructor */
		~fetch_manager() {}

		/** allocate the token */
		bool isTokenFreeForAlloc(uint32_t id, Object *obj) {
			return tokens[0].isFree();
		}

		/** this one is never used */
		bool isTokenFreeForInquire(uint32_t id, Object *obj) {
			return false;
		}

		/** blocking release - wait till cache miss served */
		bool isTokenGoodToReclaim(Token *token, Object *obj);

		/** allocate a token, start tlb/cache lookup*/ 
		Token *allocateToken(uint32_t id, Object *obj);

		/** release the token if the object is the one holding it*/
		bool reclaimToken(Token *token, Object *obj);

		void discardToken(Token *token, Object *obj);

		/** regular updates */
		void updateOnClockTick();

		/** set the program counter */
		void setProgCounter(uint32_t cnt) {progCounter = cnt;}

		uint64_t getFetchCount() {return fetch_count;}

	private:


		arm_simulator *sim;
		ICache& icache;
		ITLB& itlb;

		enum {
			FETCH_DONE,
			FETCH_TLB,		/*waiting for tlb*/
			FETCH_CACHE,	/*waiting for cache*/
		} status;

		uint64_t fetch_count;
		uint32_t progCounter;
		bool moreToCome;

} FetchManager;

/* the decoder manager can accepts all the uops spawned from the same
 * instruction simutaneously, but it makes sure that they leave
 * in order, the one with smallest dynid leaves first.
 * Note: The uops must allocate tokens with continuous and decreasing id.
 *       i.e., if there are N uops, the first should have id as N-1, the
 *       second with id N-2, ... and the last with id 0
 */
typedef class decode_manager : public Manager {
	public:

		/** constructor, maximum 16 uops from the same instruction */
		decode_manager(const std::string &name) : manager(16, name) {}

		/** destructor */
		~decode_manager() {}

		/** allocate the token */
		bool isTokenFreeForAlloc(uint32_t id, Object *obj) {
			/*must make sure that 0 is clear, the others are unimportant*/
			return tokens[0].isFree();
		}

		/** this one is never called */
		bool isTokenFreeForInquire(uint32_t id, Object *obj) {
			return false;
		}

		/** sequentialize tokens, only the largest can leave */
		bool isTokenGoodToReclaim(Token *token, Object *obj) {
			return token==tokens+maxid;
		}

		/** allocate a token*/
		Token *allocateToken(uint32_t id, Object *obj) {
			tokens[id].setUsed(obj);
			if (id>maxid) maxid = id;
			return tokens+id;
		}

		/** release the token, may be triggered by squashing */
		bool reclaimToken(Token *token, Object *obj) {
			token->setUnused();
			/*adjust maxid if necessary*/
			for (;maxid && tokens[maxid].isFree(); maxid--);
			return true;
		}

		void discardToken(Token *token, Object *obj) {
			token->setUnused();
			for (;maxid && tokens[maxid].isFree(); maxid--);
		}

	private:
		uint32_t maxid;

} DecodeManager;

typedef class buffer_manager : public Manager {

	public:

		/** constructor */
		buffer_manager(const std::string &name, DCache &cache,  DTLB& tlb) :
			manager(1, name), dcache(cache), dtlb(tlb), status(BF_DONE) {}

		/** destructor */
		~buffer_manager() {};

		/** check if empty */
		bool isTokenFreeForAlloc(uint32_t id, Object *obj) {
			return tokens[0].isFree();
		}

		/** this one is never called */
		bool isTokenFreeForInquire(uint32_t id, Object *obj) {return false;}

		/** blocking release - wait till word read */
		bool isTokenGoodToReclaim(Token *token, Object *obj);

		Token *allocateToken(uint32_t id, Object *obj);

		/** release do nothing */
		bool reclaimToken(Token *token, Object *obj) {
			token->setUnused();
			return true;
		}

        /** discard do nothing */
        void discardToken(Token *token, Object *obj);

		/** regular update */
		void updateOnClockTick();

	private:

		DCache& dcache;
		DTLB&  dtlb;

		enum {
			BF_DONE,
			BF_TLB,		/*waiting for tlb*/
			BF_CACHE,	/*waiting for cache*/
		} status;

		bool isRead, isMergeable;
		uint32_t addr;


} BFManager;

typedef class latency_manager : public Manager {

    public:

        /** constructor */
        latency_manager(const std::string &name) : manager(1, name) {}

        /** destructor */
        ~latency_manager() {};

        /** if token is free */
        bool isTokenFreeForAlloc(uint32_t id, Object *obj) {
            return tokens[0].isFree();
        }

        /** allocate a token */
        Token *allocateToken(uint32_t id, Object *obj) {
            latency = id;
            tokens[0].setUsed(obj);
            return tokens;
        }

        /** this one is never called*/
        bool isTokenFreeForInquire(uint32_t id, Object *obj) {
            return false;
        }

        /** blocking release - wait till latency satisfied */
        bool isTokenGoodToReclaim(Token *token, Object *obj) {
            return latency==0;
        }

        /** release do nothing */
        bool reclaimToken(Token *token, Object *obj) {
            token->setUnused();
            return true;
        }

        /** discard do nothing */
        void discardToken(Token *token, Object *obj) {token->setUnused();}

        /** periodical update */
        void updateOnClockTick() {
            if (latency!=0) latency--;
        }

    private:                                                  
                                                              
        uint32_t latency;                                     
                                                              
} LatencyManager;

typedef class regfile_manager : public Manager {

	public:

		/** constructor */
		regfile_manager(int n, const std::string &name, arm_simulator *sim) :
			 manager(n, name), sim(sim) {
			NullToken = new Token(this);
		}

		/** destructor */
		~regfile_manager() {
			delete NullToken;
		}

		/** allocation is preemptive */
		bool isTokenFreeForAlloc(uint32_t id, Object *obj);

		/** blocking test for data dependency modeling */
		bool isTokenFreeForInquire(uint32_t id, Object *obj);

		/** non-blocking release */
		bool isTokenGoodToReclaim(Token *token, Object *obj) {return true;}

		/** allocate a token */
		Token *allocateToken(uint32_t id, Object *obj);

		/** return a token */
		bool reclaimToken(Token *token, Object *obj);

		void discardToken(Token *token, Object *obj);

	private:

		/** the null token for those dummy register operand (immediate) */
		Token *NullToken;

		arm_simulator *sim;

} RegFileManager;

}

#endif

