/*************************************************************************
    Copyright (C) 2002,2003 Wei Qin
    See file COPYING for more information.

    This program is free software; you can redistribute it and/or modify    
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
*************************************************************************/

#include "derived_itinerary.h"
#include "manager.h"
#include "armsim.h"
#include <map>
#include <string>

typedef osm::Action::ACTION_TYPE action_type;
#define RELEASE osm::Action::ACTION_RELEASE
#define DISCARD osm::Action::ACTION_DISCARD
#define ALLOCAT	osm::Action::ACTION_ALLOCATE
#define INQUIRE	osm::Action::ACTION_INQUIRE

#define max(a,b) (a)>(b)?(a):(b)

using namespace std;
using namespace simulator;

typedef struct {
	char *src;
	char *dst;
} IGraph;

typedef struct {
	int index;
	action_type action;
	char *manager;
	int tokenInd;
	int id;
} EdgeData;

/* Dummy is the default for unimplemented instructions*/
static IGraph dummyG[] = {
	{"NU", "IF"},
	{"IF", "ID"},
	{"ID", "EX"},
	{"EX", "BF"},
	{"BF", "WB"},
	{"WB", NULL},
};

/* The pipeline stage managers are blocking in ALLOCAT,
 * non-blocking in RELEASE*/
static EdgeData dummy[] = {
	{1, RELEASE, "IF", 0, 0},
	{1, ALLOCAT, "ID", 0, 0},
	{2, DISCARD, "ID", 0, 0},
	{2, ALLOCAT, "EX", 0, 0},
	{3, DISCARD, "EX", 0, 0},
	{3, ALLOCAT, "BF", 0, 0},
	{4, DISCARD, "BF", 0, 0},
//	{4, ALLOCAT, "WB", 0, 0},
//	{5, DISCARD, "WB", 0, 0},
};

/* Two types of data processing instructions in ARM
 * 1. immediate or shift-by-immediate operand type
 * 2. shift-by-register operand type
 * Type2 takes two cycle in decoding and execution stage.
 */
static IGraph *dpi1G = dummyG;
static EdgeData dpi1[] = {
	{1, RELEASE, "IF", 0, 0},
	{1, ALLOCAT, "ID", 0, 0},
	{2, DISCARD, "ID", 0, 0},
	{2, ALLOCAT, "EX", 0, 0},
	{2, INQUIRE, "RF", 0, 0},	/*two source register operands*/
	{2, INQUIRE, "RF", 0, 0},
	{2, ALLOCAT, "RF", 1, 0},	/*destination operand*/
	{3, DISCARD, "EX", 0, 0},
	{3, ALLOCAT, "BF", 0, 0},
	{3, RELEASE, "RF", 1, 0},	/*forward back*/
	{4, DISCARD, "BF", 0, 0},
//	{4, ALLOCAT, "WB", 0, 0},
//	{5, DISCARD, "WB", 0, 0},
};

static IGraph dpi2G[] = {
	{"NU",  "IF"},
	{"IF",  "ID"},
	{"ID",  "EX1"},
	{"EX1", "EX2"},
	{"EX2", "BF"},
	{"BF",  "WB"},
	{"WB",  NULL},
};

static EdgeData dpi2[] = {
	{1, RELEASE, "IF", 0, 0},
	{1, ALLOCAT, "ID", 0, 0},
	{2, ALLOCAT, "EX", 1, 0},
	{2, INQUIRE, "RF", 0, 0},	/*three source register operands*/
	{2, INQUIRE, "RF", 0, 0},
	{2, INQUIRE, "RF", 0, 0},
	{2, ALLOCAT, "RF", 2, 0},	/*destination operand*/
	{3, DISCARD, "ID", 0, 0},	/*release ID after two cycles*/
	{4, DISCARD, "EX", 1, 0},	/*release EX after two cycles*/
	{4, ALLOCAT, "BF", 0, 0},
	{4, RELEASE, "RF", 2, 0},	/*forward back*/
	{5, DISCARD, "BF", 0, 0},
//	{5, ALLOCAT, "WB", 0, 0},
//	{6, DISCARD, "WB", 0, 0},
};

static IGraph branchG[] = {
	{"NU",  "IF"},
	{"IF",  "ID"},
	{"ID",  NULL},
};

static EdgeData branch[] = {
	{1, RELEASE, "IF", 0, 0},
	{1, ALLOCAT, "ID", 0, 0},
	{1, ALLOCAT, "RF", 1, 15},	/*register 15 is PC */
	{2, RELEASE, "ID", 0, 0},
	{2, RELEASE, "RF", 1, 15},	/*trigger control flow redirection*/
};

/*Two types of load exist in ARM,
  1. load word, load unsigned halfword/byte
  2. load signed halfword/byte.
  Type 2 takes one more cycle in BF.
 */
static IGraph *ld1G = dummyG;

static EdgeData ld1[] = {
	{1, RELEASE, "IF", 0, 0},
	{1, ALLOCAT, "ID", 0, 0},
	{2, DISCARD, "ID", 0, 0},
	{2, ALLOCAT, "EX", 0, 0},
	{2, INQUIRE, "RF", 0, 0},	/*base */
	{2, INQUIRE, "RF", 0, 0},	/*offset */
	{2, ALLOCAT, "RF", 1, 0},	/*destination*/
	{3, DISCARD, "EX", 0, 0},
	{3, ALLOCAT, "BF", 0, 0},
	{4, RELEASE, "BF", 0, 0},
	{4, RELEASE, "RF", 1, 0},	/*the loaded operand*/
//	{4, ALLOCAT, "WB", 0, 0},
//	{5, DISCARD, "WB", 0, 0},
};

static IGraph ld2G[] = {
	{"NU", "IF"},
	{"IF", "ID"},
	{"ID", "EX"},
	{"EX",  "BF1"},
	{"BF1", "BF2"},
	{"BF2", "WB"},
	{"WB", NULL},
};

static EdgeData ld2[] = {
	{1, RELEASE, "IF", 0, 0},
	{1, ALLOCAT, "ID", 0, 0},
	{2, DISCARD, "ID", 0, 0},
	{2, ALLOCAT, "EX", 0, 0},
	{2, INQUIRE, "RF", 0, 0},	/*base*/
	{2, INQUIRE, "RF", 0, 0},	/*offset*/
	{2, ALLOCAT, "RF", 1, 0},	/*destination*/
	{3, DISCARD, "EX", 0, 0},
	{3, ALLOCAT, "BF", 0, 0},
	{5, RELEASE, "BF", 0, 0},	/*release after second cycle in BF*/
	{5, RELEASE, "RF", 1, 0},	/*the loaded operand*/
//	{5, ALLOCAT, "WB", 0, 0},
//	{6, DISCARD, "WB", 0, 0},
};

static IGraph *stG = dummyG;

static EdgeData st[] = {
	{1, RELEASE, "IF", 0, 0},
	{1, ALLOCAT, "ID", 0, 0},
	{2, DISCARD, "ID", 0, 0},
	{2, ALLOCAT, "EX", 0, 0},
	{2, INQUIRE, "RF", 0, 0},	/*base*/
	{2, INQUIRE, "RF", 0, 0},	/*offset*/
	{2, INQUIRE, "RF", 0, 0},	/*to store*/
	{3, DISCARD, "EX", 0, 0},
	{3, ALLOCAT, "BF", 0, 0},
	{4, RELEASE, "BF", 0, 0},
//	{4, ALLOCAT, "WB", 0, 0},
//	{5, DISCARD, "WB", 0, 0},
};

/*filler is for ldm/stm with 0/1 register in the list or branch and link*/
static IGraph *fillerG = dummyG;

static EdgeData filler[] = {
    {1, RELEASE, "IF", 0, 0},
	{1, ALLOCAT, "ID", 0, 0},
	{2, RELEASE, "ID", 0, 0},
	{2, ALLOCAT, "EX", 0, 0},
	{2, INQUIRE, "RF", 0, 0},	/*src*/
	{2, ALLOCAT, "RF", 1, 0},	/*dest*/
	{3, DISCARD, "EX", 0, 0},
	{3, RELEASE, "RF", 1, 0},	/*dest*/
	{3, ALLOCAT, "BF", 0, 0},
	{4, RELEASE, "BF", 0, 0},
//	{4, ALLOCAT, "WB", 0, 0},
//	{5, DISCARD, "WB", 0, 0},
};

static IGraph *ldmG = dummyG;

/* a ldm is decomposed into micro-ops, in our model, all but
   the last micro-op test on IF, the last ALLOCAT IF, so all
   can be squeezed into the same IF stage.
   We donot model the destination write here since
   it is unnecessary since the consumer of the destination must
   be after ldmLast. There will be sufficient room to hide the data 
   dependency.
*/
static EdgeData ldm[] = {
    {1, RELEASE, "IF", 0, 0},
	{1, ALLOCAT, "ID", 0, 1},	/* this is special, non-exclusive allocation*/
	{2, RELEASE, "ID", 0, 0},
	{2, ALLOCAT, "EX", 0, 0},
	{2, INQUIRE, "RF", 0, 0},	/*base */
//	{2, ALLOCAT, "RF", 1, 0},	/*destination*/
	{3, DISCARD, "EX", 0, 0},
	{3, ALLOCAT, "BF", 0, 0},
	{4, RELEASE, "BF", 0, 0},
//	{3, DISCARD, "RF", 1, 0},	/*the loaded operand*/
//	{3, ALLOCAT, "WB", 0, 0},
//	{4, DISCARD, "WB", 0, 0},
};

/* from SA-110 technical reference manual
   ldm takes MAX(2, number of register loaded) cycles, so
   update of base should happen simutaneously with the last load.
   The loaded operand must be updated since it may be PC.
*/
static EdgeData ldmLast[] = {
	{1, RELEASE, "IF", 0, 0},
	{1, ALLOCAT, "ID", 0, 0},
	{2, RELEASE, "ID", 0, 0},
	{2, ALLOCAT, "EX", 0, 0},
	{2, ALLOCAT, "RF", 2, 0},	/*base */
	{2, ALLOCAT, "RF", 1, 0},	/*destination*/
	{3, DISCARD, "EX", 0, 0},
	{3, ALLOCAT, "BF", 0, 0},
	{3, RELEASE, "RF", 2, 0},	/*updated base */
	{4, RELEASE, "BF", 0, 0},
	{4, RELEASE, "RF", 1, 0},	/*the loaded operand*/
//	{4, ALLOCAT, "WB", 0, 0},
//	{5, DISCARD, "WB", 0, 0},
};

static IGraph *stmG = dummyG;

static EdgeData stm[] = {
    {1, RELEASE, "IF", 0, 0},
	{1, ALLOCAT, "ID", 0, 1},	/*1 is special, non-exclusive allocation */
	{2, RELEASE, "ID", 0, 0},
	{2, ALLOCAT, "EX", 0, 0},
	{2, INQUIRE, "RF", 0, 0},	/*base */
	{2, INQUIRE, "RF", 0, 0},	/*to store*/
	{3, DISCARD, "EX", 0, 0},
	{3, ALLOCAT, "BF", 0, 0},
	{4, RELEASE, "BF", 0, 0},
//	{4, ALLOCAT, "WB", 0, 0},
//	{5, DISCARD, "WB", 0, 0},
};

static EdgeData stmLast[] = {
	{1, RELEASE, "IF", 0, 0},
	{1, ALLOCAT, "ID", 0, 0},
	{2, RELEASE, "ID", 0, 0},
	{2, ALLOCAT, "EX", 0, 0},
	{2, ALLOCAT, "RF", 1, 0},	/*base */
	{2, INQUIRE, "RF", 0, 0},	/*to store*/
	{3, DISCARD, "EX", 0, 0},
	{3, ALLOCAT, "BF", 0, 0},
	{3, RELEASE, "RF", 1, 0},	/*updated base */
	{4, RELEASE, "BF", 0, 0},
//	{4, ALLOCAT, "WB", 0, 0},
//	{5, DISCARD, "WB", 0, 0},
};

static IGraph *mulG = dummyG;

/* if a multiplication is in EX or BF, others cannot enter EX */
static EdgeData mul[] = {
	{1, RELEASE, "IF", 0, 0},
	{1, ALLOCAT, "ID", 0, 0},
	{2, DISCARD, "ID", 0, 0},
	{2, ALLOCAT, "EX", 0, 0},
	{2, ALLOCAT, "ME", 1, 0},   /*constrains mults back to back */
	{2, INQUIRE, "RF", 0, 0},	/*src1*/
	{2, INQUIRE, "RF", 0, 0},	/*src2*/
	{2, INQUIRE, "RF", 0, 0},	/*src3*/
	{2, ALLOCAT, "RF", 2, 0},	/*dest*/
	{3, RELEASE, "EX", 0, 0},	/*release EX after # cycles*/
	{3, ALLOCAT, "BF", 0, 0},
	{4, DISCARD, "BF", 0, 0},
	{4, RELEASE, "RF", 2, 0},	/*dest*/
	{4, DISCARD, "ME", 1, 0},   /*allows next multiply to enter EX */
//	{4, DISCARD, "EX", 1, 0},	/*allows next multiply to enter EX */
//	{4, ALLOCAT, "WB", 0, 0},
//	{5, DISCARD, "WB", 0, 0},
};

static IGraph *mullG = ld2G;

static EdgeData mull[] = {
	{1, RELEASE, "IF", 0, 0},
	{1, ALLOCAT, "ID", 0, 0},
	{2, DISCARD, "ID", 0, 0},
	{2, ALLOCAT, "EX", 4, 0},
	{2, ALLOCAT, "ME", 1, 0},   /*constrains mults back to back */
	{2, INQUIRE, "RF", 0, 0},	/*src1*/
	{2, INQUIRE, "RF", 0, 0},	/*src2*/
	{2, INQUIRE, "RF", 0, 0},	/*src3 = dst1-lo*/
	{2, INQUIRE, "RF", 0, 0},	/*src4 = dst2-hi*/
	{2, ALLOCAT, "RF", 2, 0},	/*dst-lo */
	{2, ALLOCAT, "RF", 3, 0},	/*dst-hi */
	{3, ALLOCAT, "BF", 0, 0},
	{4, RELEASE, "RF", 2, 0},	/*dst-lo */
	{4, RELEASE, "EX", 4, 0},	/*release EX after # cycles*/
	{4, DISCARD, "ME", 1, 0},   /*allows next multiply to enter EX */
	{5, RELEASE, "RF", 3, 0},	/*dst-hi */
	{5, DISCARD, "BF", 0, 0},
//	{5, ALLOCAT, "WB", 0, 0},
//	{6, DISCARD, "WB", 0, 0},
};

static Token *dummyPtr;

void initItinerary(Itinerary *itin, map<string, Manager *>& managers,
	IGraph *ig, int gsize, EdgeData *edata, int esize,
	vector<Token *>& tpbuffer)
{
	int i, ntoken;
	map<string, Node *> nodes;
	map<string, Node *>::iterator nit;
	vector<Edge *> edges;

	/*create all the nodes and edges*/
	for (i=0; i<gsize; i++) {

		nit = nodes.find(ig[i].src);
		if (nit==nodes.end())
			nodes[ig[i].src] = itin->newNode(ig[i].src);

		if (ig[i].dst) {
			nit = nodes.find(ig[i].dst);
			if (nit==nodes.end())
				nodes[ig[i].dst] = itin->newNode(ig[i].dst);
			edges.push_back(itin->newEdge(nodes[ig[i].src], nodes[ig[i].dst]));
		}
		else 
			edges.push_back(itin->newEdge(nodes[ig[i].src], NULL));
	}

	/*resize token buffer*/
	for (i=0, ntoken=1; i<esize; i++)
		ntoken = max(ntoken, edata[i].tokenInd+1);
	tpbuffer.resize(ntoken);

	/*create all actions*/
	for (i=0; i<esize; i++) {
		assert(edata[i].index<gsize);
		assert(managers.find(edata[i].manager)!=managers.end());
		/* since in this implementation we don't use the token returned     
		by inquire, we give the action a dummy ptr to place the token*/  
        Token **pptr = (edata[i].action==INQUIRE)?(&dummyPtr):
            (&tpbuffer[edata[i].tokenInd]);                   
        Action *action = new Action(edata[i].action,          
            managers[edata[i].manager], pptr, edata[i].id);
		edges[edata[i].index]->addAction(action);
	}
}

dummy_itinerary::dummy_itinerary(map<string, Manager *>& managers) :
	itinerary(ITN_DUMMY)
{
	int gsize = sizeof(dummyG)/sizeof(IGraph);
	int esize = sizeof(dummy)/sizeof(EdgeData);
	initItinerary(this, managers, dummyG, gsize, dummy, esize, tpbuffer);
}

void dummy_itinerary::setIds(uint32_t ids[])
{
}

dpi1_itinerary::dpi1_itinerary(map<string, Manager *>& managers) :
	itinerary (ITN_DPI1)
{
	int gsize = sizeof(dummyG)/sizeof(IGraph);
	int esize = sizeof(dpi1)/sizeof(EdgeData);
	initItinerary(this, managers, dpi1G, gsize, dpi1, esize, tpbuffer);
}

void dpi1_itinerary::setIds(uint32_t ids[])
{
	edges[2]->setActionID(2, ids[0]);/*src1*/
	edges[2]->setActionID(3, ids[1]);/*src2*/
	edges[2]->setActionID(4, ids[2]);/*dst*/
	edges[3]->setActionID(2, ids[2]);/*dst*/
}

dpi2_itinerary::dpi2_itinerary(map<string, Manager *>& managers) :
	itinerary(ITN_DPI2)
{
	int gsize = sizeof(dpi2G)/sizeof(IGraph);
	int esize = sizeof(dpi2)/sizeof(EdgeData);
	initItinerary(this, managers, dpi2G, gsize, dpi2, esize, tpbuffer);
}

void dpi2_itinerary::setIds(uint32_t ids[])
{
	edges[2]->setActionID(1, ids[0]);/*src1*/
	edges[2]->setActionID(2, ids[1]);/*src2*/
	edges[2]->setActionID(3, ids[2]);/*src3*/
	edges[2]->setActionID(4, ids[3]);/*dst*/
	edges[4]->setActionID(2, ids[3]);/*dst*/
}

/*branch & branch_link*/
branch_itinerary::branch_itinerary(map<string, Manager *>& managers) :
	itinerary(ITN_BRANCH)
{
	int gsize = sizeof(branchG)/sizeof(IGraph);
	int esize = sizeof(branch)/sizeof(EdgeData);
	initItinerary(this, managers, branchG, gsize, branch, esize, tpbuffer);
}

void branch_itinerary::setIds(uint32_t ids[])
{
}

/*
  Since the specification states that if R15 is used as RN and
  updated the result is undefined,
  we can ignore the updat of RN here.
  It's single-cycle latency is take care of by EX.
*/
ld1_itinerary::ld1_itinerary(map<string, Manager *>& managers) :
	itinerary(ITN_LD1)
{
	int gsize = sizeof(dummyG)/sizeof(IGraph);
	int esize = sizeof(ld1)/sizeof(EdgeData);
	initItinerary(this, managers, ld1G, gsize, ld1, esize, tpbuffer);
}

void ld1_itinerary::setIds(uint32_t ids[])
{
	edges[2]->setActionID(2, ids[0]);/*base*/
	edges[2]->setActionID(3, ids[1]);/*offset*/
	edges[2]->setActionID(4, ids[2]);/*dst*/
	edges[3]->setActionID(1, ids[3]);/*addr*/
	edges[4]->setActionID(1, ids[2]);/*dst*/
}

ld2_itinerary::ld2_itinerary(map<string, Manager *>& managers) :
	itinerary(ITN_LD2)
{
	int gsize = sizeof(ld2G)/sizeof(IGraph);
	int esize = sizeof(ld2)/sizeof(EdgeData);
	initItinerary(this, managers, ld2G, gsize, ld2, esize, tpbuffer);
}

void ld2_itinerary::setIds(uint32_t ids[])
{
	edges[2]->setActionID(2, ids[0]);/*base*/
	edges[2]->setActionID(3, ids[1]);/*offset*/
	edges[2]->setActionID(4, ids[2]);/*dst*/
	edges[3]->setActionID(1, ids[3]);/*addr*/
	edges[5]->setActionID(1, ids[2]);/*dst*/
}

st_itinerary::st_itinerary(map<string, Manager *>& managers) :
	itinerary(ITN_ST)
{
	int gsize = sizeof(dummyG)/sizeof(IGraph);
	int esize = sizeof(st)/sizeof(EdgeData);
	initItinerary(this, managers, stG, gsize, st, esize, tpbuffer);
}

void st_itinerary::setIds(uint32_t ids[])
{
	edges[2]->setActionID(2, ids[0]);/*base*/
	edges[2]->setActionID(3, ids[1]);/*offset*/
	edges[2]->setActionID(4, ids[2]);/*tostore*/
	edges[3]->setActionID(1, ids[3]);/*addr*/
}

filler_itinerary::filler_itinerary(map<string, Manager *>& managers) :
	itinerary(ITN_FILLER)
{
	int gsize = sizeof(dummyG)/sizeof(IGraph);
	int esize = sizeof(filler)/sizeof(EdgeData);
	initItinerary(this, managers, fillerG, gsize, filler, esize, tpbuffer);
}

void filler_itinerary::setIds(uint32_t ids[])
{
	edges[2]->setActionID(2, ids[0]);/*src*/
	edges[2]->setActionID(3, ids[1]);/*dest*/
	edges[3]->setActionID(1, ids[1]);/*dest*/
	edges[1]->setActionID(1, ids[2]);/*the decoder id*/
}

ldm_itinerary::ldm_itinerary(map<string, Manager *>& managers) :
	itinerary(ITN_LDM)
{
	int gsize = sizeof(dummyG)/sizeof(IGraph);
	int esize = sizeof(ldm)/sizeof(EdgeData);
	initItinerary(this, managers, ldmG, gsize, ldm, esize, tpbuffer);
}

void ldm_itinerary::setIds(uint32_t ids[])
{
	edges[2]->setActionID(2, ids[0]);/*base*/
//	edges[2]->setActionID(3, ids[1]);/*dest*/
	edges[3]->setActionID(1, ids[2]);/*addr*/
//	edges[4]->setActionID(1, ids[1]);/*dest*/
	edges[1]->setActionID(1, ids[3]);/*index*/
}

ldmLast_itinerary::ldmLast_itinerary(map<string, Manager *>& managers) :
	itinerary(ITN_LDMLAST)
{
	int gsize = sizeof(dummyG)/sizeof(IGraph);
	int esize = sizeof(ldmLast)/sizeof(EdgeData);
	initItinerary(this, managers, dummyG, gsize, ldmLast, esize, tpbuffer);
}

void ldmLast_itinerary::setIds(uint32_t ids[])
{
	edges[2]->setActionID(2, ids[0]);/*base*/
	edges[2]->setActionID(3, ids[1]);/*dest*/
	edges[3]->setActionID(2, ids[0]);/*new base*/
	edges[3]->setActionID(1, ids[2]);/*addr*/
	edges[4]->setActionID(1, ids[1]);/*dest*/
}

stm_itinerary::stm_itinerary(map<string, Manager *>& managers) :
	itinerary(ITN_STM)
{
	int gsize = sizeof(dummyG)/sizeof(IGraph);
	int esize = sizeof(stm)/sizeof(EdgeData);
	initItinerary(this, managers, stmG, gsize, stm, esize, tpbuffer);
}

void stm_itinerary::setIds(uint32_t ids[])
{
	edges[2]->setActionID(2, ids[0]);/*base*/
	edges[2]->setActionID(3, ids[1]);/*tostore*/
	edges[3]->setActionID(1, ids[2]);/*addr*/
	edges[1]->setActionID(1, ids[3]);/*index*/
}

stmLast_itinerary::stmLast_itinerary(map<string, Manager *>& managers) :
	itinerary(ITN_STMLAST)
{
	int gsize = sizeof(dummyG)/sizeof(IGraph);
	int esize = sizeof(stmLast)/sizeof(EdgeData);
	initItinerary(this, managers, dummyG, gsize, stmLast, esize, tpbuffer);
}

void stmLast_itinerary::setIds(uint32_t ids[])
{
	edges[2]->setActionID(2, ids[0]);/*base*/
	edges[2]->setActionID(3, ids[1]);/*tostore*/
	edges[3]->setActionID(2, ids[0]);/*new base*/
	edges[3]->setActionID(1, ids[2]);/*addr*/
}

mul_itinerary::mul_itinerary(map<string, Manager *>& managers) :
	itinerary(ITN_MUL)
{
	int gsize = sizeof(dummyG)/sizeof(IGraph);
	int esize = sizeof(mul)/sizeof(EdgeData);
	initItinerary(this, managers, mulG, gsize, mul, esize, tpbuffer);
}

void mul_itinerary::setIds(uint32_t ids[])
{
	edges[2]->setActionID(3, ids[0]);/*src1*/
	edges[2]->setActionID(4, ids[1]);/*src2*/
	edges[2]->setActionID(5, ids[2]);/*src3*/
	edges[2]->setActionID(6, ids[3]);/*dst*/
	edges[2]->setActionID(1, ids[4]);/*count*/
	edges[4]->setActionID(1, ids[3]);/*dst*/
}

mull_itinerary::mull_itinerary(map<string, Manager *>& managers) :
	itinerary(ITN_MULL)
{
	int gsize = sizeof(ld2G)/sizeof(IGraph);
	int esize = sizeof(mull)/sizeof(EdgeData);
	initItinerary(this, managers, mullG, gsize, mull, esize, tpbuffer);
}

void mull_itinerary::setIds(uint32_t ids[])
{
	edges[2]->setActionID(3, ids[0]);/*src1*/
	edges[2]->setActionID(4, ids[1]);/*src2*/
	edges[2]->setActionID(5, ids[2]);/*src3=dst-lo*/
	edges[2]->setActionID(6, ids[3]);/*src4=dst-hi*/
	edges[2]->setActionID(7, ids[4]);/*dst-lo*/
	edges[2]->setActionID(8, ids[5]);/*dst-hi*/
	edges[2]->setActionID(1, ids[6]);/*count*/
	edges[4]->setActionID(0, ids[4]);/*dst-lo*/
	edges[5]->setActionID(0, ids[5]);/*dst-hi*/
}
