/*************************************************************************
    Copyright (C) 2002,2003 Wei Qin
    See file COPYING for more information.

    This program is free software; you can redistribute it and/or modify    
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
*************************************************************************/

#include "armsim.h"
#include "ldst_itin.h"
#include "derived_itinerary.h"
#include "itinerary_pool.h"

/*memory opcode fields*/
#define BFLD ((inst>>22)&1)
#define UFLD ((inst>>23)&1)
#define PFLD ((inst>>24)&1)
#define WFLD ((inst>>21)&1)
#define SIGN ((inst>>6)&1)
#define HFLD ((inst>>5)&1)
#define LFLD ((inst>>20)&1)

#undef WRITE_REG
#undef READ_REG
#undef READ_CPSR

#define WRITE_REG(ind, val) (sim->emu.write_gpr(ind,val))
#define READ_REG(ind) (sim->emu.read_gpr(ind))
#define READ_CPSR  (sim->emu.read_cpsr())

using namespace simulator;

static inline UInt32 popcount16(UInt32 x)
{
  x &= 0xFFFF;
  x = (x & 0x5555) + ((x >> 1) & 0x5555);
  x = (x & 0x3333) + ((x >> 2) & 0x3333);
  x = (x & 0x0F0F) + ((x >> 4) & 0x0F0F);
  x = (x & 0x00FF) + ((x >> 8) & 0x00FF);
  return x;
}

static inline UInt32 lmb16(UInt32 x)
{
  UInt32 i;
  x &= 0xFFFF;
  for (i=15; i>0; i--)
	if (x&(1<<i)) return i;
  return x-1;
}

static inline UInt32 rmb16(UInt32 x)
{
  UInt32 i;
  x &= 0xFFFF;
  for (i=0; i<16; i++)
	if (x&(1<<i)) return i;
  return (UInt32)-1;
}

static inline UInt32 rotate_right(UInt32 val, UInt32 imm)
{
	return (val >> imm ) | (val << (32 - imm));
}

/*shifter operand*/
static inline UInt32 shifter_operand(arm_simulator *sim, arm_inst_t inst)
{
	UInt32 shift_imm = (inst>>7) & 0x01f, val = RM, result;
	UInt8 type = (inst>>5) & 0x03;

	switch (type) {
		case 0: /*LSL*/
			result = val << shift_imm;
			break;
		case 1: /*LSR*/
			if (shift_imm) {
				result = val >> shift_imm;
			} else {
				result = 0;
			}
			break; 
		case 2: /*ASR*/ 
			if (shift_imm) {
				result = (SInt32)val>>shift_imm;
			} 
			else {
				result = BIT31(val)?~0:BIT31(val);
			}
			break; 
		default: /*ROR*/
			if (shift_imm) {
				result = rotate_right(val, shift_imm);
			}
			else { /*RRX*/									
				result = (val>>1) | (C_FLAG<<31);			 
			}												 
			break;											
	}														 
	return result;											
}

/*P==0, B==0*/
Itinerary *itn_ld1_imm(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t ids[4] = {RNFLD, (uint32_t)-1, RDFLD, RN};
	Itinerary *itn = sim->itnq->allocate(ITN_LD1);
	itn->setIds(ids);
	return itn;
}

/*P==1, B==0*/
Itinerary *itn_ld1_imm_p(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t offset = inst & 0xFFF, address = UFLD?RN+offset:RN-offset;  

	uint32_t ids[4] = {RNFLD, (uint32_t)-1, RDFLD, address};
	Itinerary *itn = sim->itnq->allocate(ITN_LD1);
	itn->setIds(ids);
	return itn;
}

/*P==0, B==1*/
Itinerary *itn_ld1_imm_b(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t ids[4] = {RNFLD, (uint32_t)-1, RDFLD, RN};
	Itinerary *itn = sim->itnq->allocate(ITN_LD1);
	itn->setIds(ids);
	return itn;
}

/*P==1, B==1*/
Itinerary *itn_ld1_imm_pb(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t offset = inst & 0xFFF, address = UFLD?RN+offset:RN-offset;  
	uint32_t ids[4] = {RNFLD, (uint32_t)-1, RDFLD, address};
	Itinerary *itn = sim->itnq->allocate(ITN_LD1);
	itn->setIds(ids);
	return itn;
}

/*p==0, b==0*/
Itinerary *itn_ld1_reg(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t ids[4] = {RNFLD, RMFLD, RDFLD, RN};
	Itinerary *itn = sim->itnq->allocate(ITN_LD1);
	itn->setIds(ids);
	return itn;
}

/*p==1, b==0*/
Itinerary *itn_ld1_reg_p(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t offset = shifter_operand(sim, inst),
   		address = UFLD?RN+offset:RN-offset;  
	uint32_t ids[4] = {RNFLD, RMFLD, RDFLD, address};
	Itinerary *itn = sim->itnq->allocate(ITN_LD1);
	itn->setIds(ids);
	return itn;
}

/*p==0, b==1*/
Itinerary *itn_ld1_reg_b(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t ids[4] = {RNFLD, RMFLD, RDFLD, RN};
	Itinerary *itn = sim->itnq->allocate(ITN_LD1);
	itn->setIds(ids);
	return itn;
}

/*p==1, b==1*/
Itinerary *itn_ld1_reg_pb(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t offset = shifter_operand(sim, inst),
   		address = UFLD?RN+offset:RN-offset;  
	uint32_t ids[4] = {RNFLD, RMFLD, RDFLD, address};
	Itinerary *itn = sim->itnq->allocate(ITN_LD1);
	itn->setIds(ids);
	return itn;
}

Itinerary *itn_ld2_imm(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t offset = ((inst>>4) & 0xF) | (inst & 0xF);		 
	uint32_t address;
	if (PFLD) /*preindex*/
		address = UFLD?RN+offset:RN-offset;
	else
		address = RN;
	uint32_t ids[4] = {RNFLD, (uint32_t)-1, RDFLD, address};
	if (!HFLD || SIGN) {
		Itinerary *itn = sim->itnq->allocate(ITN_LD2);
		itn->setIds(ids);
		return itn;
	}
	else {
		Itinerary *itn = sim->itnq->allocate(ITN_LD1);
		itn->setIds(ids);
		return itn;
	}
}

Itinerary *itn_ld2_reg(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t offset = RM;
	uint32_t address;										   
	if (PFLD) /*preindex*/
		address = UFLD?RN+offset:RN-offset;				   
	else													  
		address = RN;   
	uint32_t ids[4] = {RNFLD, RMFLD, RDFLD, address};
	if (!HFLD || SIGN) {
		Itinerary *itn = sim->itnq->allocate(ITN_LD2);
		itn->setIds(ids);
		return itn;
	}
	else {
		Itinerary *itn = sim->itnq->allocate(ITN_LD1);
		itn->setIds(ids);
		return itn;
	}
}

/*p==0, b==0*/
Itinerary *itn_st1_imm(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t ids[4] = {RNFLD, (uint32_t)-1, RDFLD, RN};
	Itinerary *itn = sim->itnq->allocate(ITN_ST);
	itn->setIds(ids);
	return itn;
}

/*p==1, b==0*/
Itinerary *itn_st1_imm_p(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t offset = inst & 0xFFF, address = UFLD?RN+offset:RN-offset;  
	uint32_t ids[4] = {RNFLD, (uint32_t)-1, RDFLD, address};
	Itinerary *itn = sim->itnq->allocate(ITN_ST);
	itn->setIds(ids);
	return itn;
}

/*p==0, b==1*/
Itinerary *itn_st1_imm_b(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t ids[4] = {RNFLD, (uint32_t)-1, RDFLD, RN};
	Itinerary *itn = sim->itnq->allocate(ITN_ST);
	itn->setIds(ids);
	return itn;
}

/*p==1, b==1*/
Itinerary *itn_st1_imm_pb(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t offset = inst & 0xFFF, address = UFLD?RN+offset:RN-offset;  
	uint32_t ids[4] = {RNFLD, (uint32_t)-1, RDFLD, address};
	Itinerary *itn = sim->itnq->allocate(ITN_ST);
	itn->setIds(ids);
	return itn;
}

/*p==0, b==0*/
Itinerary *itn_st1_reg(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t ids[4] = {RNFLD, RMFLD, RDFLD, RN};
	Itinerary *itn = sim->itnq->allocate(ITN_ST);
	itn->setIds(ids);
	return itn;
}

/*p==1, b==0*/
Itinerary *itn_st1_reg_p(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t offset = shifter_operand(sim, inst),
   		address = UFLD?RN+offset:RN-offset;  
	uint32_t ids[4] = {RNFLD, RMFLD, RDFLD, address};
	Itinerary *itn = sim->itnq->allocate(ITN_ST);
	itn->setIds(ids);
	return itn;
}

/*p==0, b==1*/
Itinerary *itn_st1_reg_b(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t ids[4] = {RNFLD, RMFLD, RDFLD, RN};
	Itinerary *itn = sim->itnq->allocate(ITN_ST);
	itn->setIds(ids);
	return itn;
}

/*p==1, b==1*/
Itinerary *itn_st1_reg_pb(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t offset = shifter_operand(sim, inst),
   		address = UFLD?RN+offset:RN-offset;  
	uint32_t ids[4] = {RNFLD, RMFLD, RDFLD, address};
	Itinerary *itn = sim->itnq->allocate(ITN_ST);
	itn->setIds(ids);
	return itn;
}

Itinerary *itn_st2_imm(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t offset = ((inst>>4) & 0xF) | (inst & 0xF);		 
	uint32_t address;
	if (PFLD) /*preindex*/
		address = UFLD?RN+offset:RN-offset;
	else
		address = RN;
	uint32_t ids[4] = {RNFLD, (uint32_t)-1, RDFLD, address};
	Itinerary *itn = sim->itnq->allocate(ITN_ST);
	itn->setIds(ids);
	return itn;
}

Itinerary *itn_st2_reg(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t offset = RM;
	uint32_t address;										   
	if (PFLD) /*preindex*/
		address = UFLD?RN+offset:RN-offset;				   
	else													  
		address = RN;   
	uint32_t ids[4] = {RNFLD, (uint32_t)-1, RDFLD, address};
	Itinerary *itn = sim->itnq->allocate(ITN_ST);
	itn->setIds(ids);
	return itn;
}

/*TODO: Implement the real itinerary for swaps.
  These two are VERY rarely used by gcc. */
Itinerary *itn_swap(arm_simulator *sim, arm_inst_t inst)
{
	Itinerary *itn = sim->itnq->allocate(ITN_DUMMY);
	return itn;
}

Itinerary *itn_swap_byte(arm_simulator *sim, arm_inst_t inst)
{
	Itinerary *itn = sim->itnq->allocate(ITN_DUMMY);
	return itn;
}


#define UMINUS1 ((uint32_t)-1)

/*simplified address RN */
Itinerary *itn_ldm(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t start_addr, end_addr, address;
	int i, lmb = lmb16(inst), ucount = popcount16(inst), pcount = ucount<<2;

	if (!PFLD && UFLD) {
		start_addr = RN;
		end_addr = RN + pcount - 4;
	}
	else if (PFLD && UFLD) {
		start_addr = RN+4;
		end_addr = RN + pcount;
	}
	else if (!PFLD && !UFLD) {
		start_addr = RN - pcount + 4;
		end_addr = RN;
	}
	else {
		start_addr = RN - pcount;
		end_addr = RN - 4;
	}

	/* 0 register load, use 2 filler itineraries*/
	if (ucount==0) {
		sim->uopBufferSize = 2;
		/*need RN?? */
		uint32_t ids1[] = {RNFLD, UMINUS1, 1};
		Itinerary *itn = sim->itnq->allocate(ITN_FILLER);
		itn->setIds(ids1);
		sim->uopBuffer[1] = itn;

		/*this one updates RN if WFLD*/
		uint32_t ids2[] = {UMINUS1, WFLD?RNFLD:UMINUS1, 0};
		itn = sim->itnq->allocate(ITN_FILLER);
		itn->setIds(ids2);
		sim->uopBuffer[0] = itn;
	}
	else if (ucount==1) {
		sim->uopBufferSize = 2;
		/* one ldm first */
		address = start_addr - (start_addr & 0x3);
		uint32_t ids1[] = {RNFLD, lmb, address, 1};
		Itinerary *itn = sim->itnq->allocate(ITN_LDM);
		itn->setIds(ids1);
		sim->uopBuffer[1] = itn;

		/*then one filler, updates RN if WFLD*/
		uint32_t ids2[] = {UMINUS1, WFLD?RNFLD:UMINUS1, 0};
		itn = sim->itnq->allocate(ITN_FILLER);
		itn->setIds(ids2);
		sim->uopBuffer[0] = itn;
	}
	else {
		bool first=true;
		sim->uopBufferSize = ucount;

		address = start_addr - (start_addr & 0x3);
		for (i=0; i<lmb; i++) {
			if ((1<<i) & inst) {
				/* only the first will test on RNFLD*/
				uint32_t ids[] = {first?RNFLD:UMINUS1, i, address, --ucount};
				Itinerary *itn = sim->itnq->allocate(ITN_LDM);
				itn->setIds(ids);
				sim->uopBuffer[ucount] = itn;
				address += 4;
				first = false;
			}
		}
		assert(ucount==1);

		uint32_t ids[] = {WFLD?RNFLD:UMINUS1, lmb, address};
		Itinerary *itn = sim->itnq->allocate(ITN_LDMLAST);
		itn->setIds(ids);
		sim->uopBuffer[0] = itn;
		//assert(end_addr == address);
	}

	Itinerary *itn = sim->uopBuffer[--sim->uopBufferSize];
	return itn;
}

Itinerary *itn_stm(arm_simulator *sim, arm_inst_t inst)
{
	uint32_t start_addr, end_addr, address;
	int i, lmb = lmb16(inst), ucount = popcount16(inst), pcount = ucount<<2;

	if (!PFLD && UFLD) {
		start_addr = RN;
		end_addr = RN + pcount - 4;
	}
	else if (PFLD && UFLD) {
		start_addr = RN+4;
		end_addr = RN + pcount;
	}
	else if (!PFLD && !UFLD) {
		start_addr = RN - pcount + 4;
		end_addr = RN;
	}
	else {
		start_addr = RN - pcount;
		end_addr = RN - 4;
	}

	/* 0 register store, use two fillers*/
	if (ucount==0) {
		sim->uopBufferSize = 2;
		/*need RN?? */
		uint32_t ids1[] = {RNFLD, UMINUS1, 1};
		Itinerary *itn = sim->itnq->allocate(ITN_FILLER);
		itn->setIds(ids1);
		sim->uopBuffer[1] = itn;

		/*this one updates RN if WFLD*/
		uint32_t ids2[] = {UMINUS1, WFLD?RNFLD:UMINUS1, 0};
		itn = sim->itnq->allocate(ITN_FILLER);
		itn->setIds(ids2);
		sim->uopBuffer[0] = itn;
	}
	else if (ucount==1) {
		sim->uopBufferSize = 2;
		/* one stm */
		address = start_addr - (start_addr & 0x3);
		uint32_t ids1[] = {RNFLD, lmb, address, 1};
		Itinerary *itn = sim->itnq->allocate(ITN_STM);
		itn->setIds(ids1);
		sim->uopBuffer[1] = itn;

		/*then one filler, updates RN if WFLD*/
		uint32_t ids2[] = {UMINUS1, WFLD?RNFLD:UMINUS1, 0};
		itn = sim->itnq->allocate(ITN_FILLER);
		itn->setIds(ids2);
		sim->uopBuffer[0] = itn;
	}
	else {
		bool first=true;
		sim->uopBufferSize = ucount;

		address = start_addr - (start_addr & 0x3);
		for (i=0; i<lmb; i++) {
			if ((1<<i) & inst) {
				uint32_t ids[] = {first?RNFLD:UMINUS1, i, address, --ucount};
				Itinerary *itn = sim->itnq->allocate(ITN_STM);
				itn->setIds(ids);
				sim->uopBuffer[ucount] = itn;
				address += 4;
				first = false;
			}
		}
		assert(ucount==1);

		uint32_t ids[] = {WFLD?RNFLD:UMINUS1, lmb, address};
		Itinerary *itn = sim->itnq->allocate(ITN_STMLAST);
		itn->setIds(ids);
		sim->uopBuffer[0] = itn;
		//assert(end_addr == address);
	}

	Itinerary *itn = sim->uopBuffer[--sim->uopBufferSize];
	return itn;
}

